""" MultiQC module to parse output from BISCUITqc """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph, beeswarm

# Initialize the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Inherits from base module class
    Initialize data structures and prepare BISCUIT report
    Inputs:
        No inputs
    Returns:
        BISCUIT report for MultiQC
    """

    def __init__(self):
        # Initialize the parent object
        super(MultiqcModule, self).__init__(
            name="BISCUIT",
            anchor="biscuit",
            href="https://github.com/huishenlab/biscuit",
            info="is a tool to map bisulfite converted DNA sequence reads and determine cytosine methylation states.",
            # Can't find a DOI // doi=
        )

        # Set up data structures
        self.mdata = {
            # General statistics
            "align_mapq": {},
            "align_strand": {},
            "align_isize": {},
            # Duplicate reporting
            "dup_report": {},
            # Uniformity
            "qc_cv": {},
            # Base coverage
            "covdist_all_base_botgc": {},
            "covdist_all_base": {},
            "covdist_all_base_topgc": {},
            "covdist_q40_base_botgc": {},
            "covdist_q40_base": {},
            "covdist_q40_base_topgc": {},
            # CpG coverage
            "covdist_all_cpg_botgc": {},
            "covdist_all_cpg": {},
            "covdist_all_cpg_topgc": {},
            "covdist_q40_cpg_botgc": {},
            "covdist_q40_cpg": {},
            "covdist_q40_cpg_topgc": {},
            # Cytosine retention
            "cpg_retention_readpos": {},
            "cph_retention_readpos": {},
            "base_avg_retention_rate": {},
            "read_avg_retention_rate": {},
        }

        # NB: Cleaning filenames like this means that some MultiQC functionality, like -s / --fullnames doesn't work.
        # However, because some of the parsing relies on cleaned filenames, the above options break the
        # module if we use the centralised MultiQC functions.
        file_suffixes = [
            # General statistics
            ".txt",
            "_mapq_table",
            "_strand_table",
            "_isize_table",
            # Duplicate reporting
            "_dup_report",
            # Uniformity
            "_cv_table",
            # Base coverage
            "_covdist_all_base_botgc_table",
            "_covdist_all_base_table",
            "_covdist_all_base_topgc_table",
            "_covdist_q40_base_botgc_table",
            "_covdist_q40_base_table",
            "_covdist_q40_base_topgc_table",
            # CpG coverage
            "_covdist_all_cpg_botgc_table",
            "_covdist_all_cpg_table",
            "_covdist_all_cpg_topgc_table",
            "_covdist_q40_cpg_botgc_table",
            "_covdist_q40_cpg_table",
            "_covdist_q40_cpg_topgc_table",
            # Cytosine retention
            "_CpGRetentionByReadPos",
            "_CpHRetentionByReadPos",
            "_totalBaseConversionRate",
            "_totalReadConversionRate",
        ]

        # Find and parse alignment reports
        for k in self.mdata:
            for f in self.find_log_files(f"biscuit/{k}"):
                s_name = f["fn"]
                for suffix in file_suffixes:
                    s_name = s_name.replace(suffix, "")
                s_name = self.clean_s_name(s_name, f)

                # Add source file to multiqc_sources.txt
                self.add_data_source(f, s_name=s_name, section=k)

                if s_name in self.mdata[k]:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")

                self.mdata[k][s_name] = getattr(self, f"parse_logs_{k}")(f["f"], f["fn"])

        for k in self.mdata:
            self.mdata[k] = self.ignore_samples(self.mdata[k])

        n_samples = max([len(self.mdata[k]) for k in self.mdata])

        if n_samples == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {n_samples} samples")

        # Basic stats table
        self.biscuit_stats_table()

        # Write data to file
        self.write_data_file(self.mdata, "biscuit")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Make report sections
        for k in self.mdata:
            if len(self.mdata[k]) > 0:
                log.debug(f"Found {len(self.mdata[k])} {k} reports")
                getattr(self, f"chart_{k}")()

    def biscuit_stats_table(self):
        """
        Create general statistics table for BISCUIT data
        Inputs:
            Uses mdata['align_mapq'] and mdata['dup_report']
        Returns:
            Add columns to MultiQC general statistics table
        """
        pd = {}

        # Calculate % aligned
        for s_name, dd in self.mdata["align_mapq"].items():
            if len(dd) > 0:
                pd[s_name] = {"aligned": dd["frc_align"]}

        # Calculate % duplicated
        for s_name, dd in self.mdata["dup_report"].items():
            if s_name not in pd:
                pd[s_name] = {}
            if "all" in dd and dd["all"] != -1:
                pd[s_name]["dup_all"] = dd["all"]
            if "q40" in dd and dd["q40"] != -1:
                pd[s_name]["dup_q40"] = dd["q40"]

        pheader = {
            "dup_q40": {
                "title": "Dup. % for Q40 Reads",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "YlOrBr",
                "hidden": True,
            },
            "dup_all": {"title": "Dup. % for All Reads", "min": 0, "max": 100, "suffix": "%", "scale": "Reds"},
            "aligned": {
                "title": "% Aligned",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
                "format": "{:,.2f}",
            },
        }

        self.general_stats_addcols(pd, pheader)

    ########################################
    #####  General Mapping Information #####
    ########################################
    @staticmethod
    def parse_logs_align_mapq(f, fn):
        """
        Parse _mapq_table.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of aligned mapq data
        """
        file_data = f.splitlines()[2:]

        # Handle missing data
        if len(file_data) == 0:
            log.debug(f"No data available in {fn}. Will not fill corresponding entries.")
            return {}

        mapq = {}
        for line in file_data:
            s = line.split()
            mapq[s[0]] = s[1]  # mapq[MAPQ] = number of reads

        data = {
            "frc_align": 0,
            "opt_align": 0,
            "sub_align": 0,
            "not_align": 0,
            "mapqs": dict(zip(range(61), [0 for _ in range(61)])),
        }
        if len(mapq) > 0:
            total = sum([int(cnt) for _, cnt in mapq.items() if _ != "unmapped"])
            for mq, cnt in mapq.items():
                if mq == "unmapped":
                    data["not_align"] += int(cnt)
                else:
                    data["mapqs"][int(mq)] = 100.0 * float(cnt) / total
                    if int(mq) >= 40:
                        data["opt_align"] += int(cnt)
                    else:
                        data["sub_align"] += int(cnt)

            data["frc_align"] = (
                100
                * (data["opt_align"] + data["sub_align"])
                / (data["opt_align"] + data["sub_align"] + data["not_align"])
            )

        return data

    ########################################
    ####    Alignment Quality Report    ####
    ########################################
    def chart_align_mapq(self):
        """
        Chart _mapq_table.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Mapping Overview and Mapping Quality
            Distribution charts
        """

        #
        # Mapping Overview bar chart
        #
        pd = {}

        # Calculate alignment counts
        for s_name, dd in self.mdata["align_mapq"].items():
            if len(dd) > 0:
                pd[s_name] = {"opt_align": dd["opt_align"], "sub_align": dd["sub_align"], "not_align": dd["not_align"]}

        pheader = {
            "opt_align": {"color": "#1f78b4", "name": "Optimally Aligned Reads"},
            "sub_align": {"color": "#a6cee3", "name": "Suboptimally Aligned Reads"},
            "not_align": {"color": "#b2df8a", "name": "Unaligned Reads"},
        }

        pconfig = {
            "id": "biscuit-mapping-overview-plot",
            "title": "BISCUIT: Mapping Overview",
            "ylab": "Number of Reads",
            "cpswitch_counts_label": "# Reads",
        }

        self.add_section(
            name="Mapping Overview",
            anchor="biscuit-mapping-overview",
            description="""
                Number of optimally aligned reads (`MAPQ>=40`), suboptimally
                aligned reads (`MAPQ<40`), and unmapped reads. Primary alignments only.
            """,
            helptext="""
                A good library should have a high fraction of reads
                that are optimally aligned. Note, suboptimally aligned reads
                include both non-unique alignments and imperfect alignments.
            """,
            plot=bargraph.plot(pd, pheader, pconfig),
        )

        #
        # Mapping Quality Distribution line graph
        #

        # Calculate the % aligned for each mapping q score
        pd_mapq = {}
        for s_name, dd in self.mdata["align_mapq"].items():
            if len(dd) > 0:
                pd_mapq[s_name] = dd["mapqs"]

        pconfig = {
            "id": "biscuit_mapq",
            "title": "BISCUIT: Distribution of Mapping Qualities",
            "ymin": 0,
            "xmin": 0,
            "yLabelFormat": "{value}%",
            "tt_label": "<strong>Q{point.x}:</strong> {point.y:.2f}% of mapped reads",
            "ylab": "% of Primary Mapped Reads",
            "xlab": "Mapping Quality Score",
        }

        self.add_section(
            name="Mapping Quality Distribution",
            anchor="biscuit-mapq",
            description="""
                The percentage of the total number of mapped reads
                for each mapping quality score. Primary alignments only.
            """,
            helptext="""
                A good quality sample should have a high quality mapping score
                for the majority of alignments.
            """,
            plot=linegraph.plot(pd_mapq, pconfig),
        )

    ########################################
    ####     Strand Alignment Report    ####
    ########################################
    def parse_logs_align_strand(self, f, fn):
        """
        Parse _strand_table.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of strand data for reads 1 and 2
        """
        patterns = [
            r"(R1)\s+\((f)\)\:\s+(\d+)\s+(\d+)",
            r"(R1)\s+\((r)\)\:\s+(\d+)\s+(\d+)",
            r"(R2)\s+\((f)\)\:\s+(\d+)\s+(\d+)",
            r"(R2)\s+\((r)\)\:\s+(\d+)\s+(\d+)",
        ]

        data = {"read1": {}, "read2": {}}
        for pat in patterns:
            m = re.search(pat, f, re.MULTILINE)
            if m is not None:
                if m.group(1) == "R1":
                    if m.group(2) == "f":
                        data["read1"]["ff"] = int(m.group(3))
                        data["read1"]["fr"] = int(m.group(4))
                    else:
                        data["read1"]["rf"] = int(m.group(3))
                        data["read1"]["rr"] = int(m.group(4))
                else:
                    if m.group(2) == "f":
                        data["read2"]["ff"] = int(m.group(3))
                        data["read2"]["fr"] = int(m.group(4))
                    else:
                        data["read2"]["rf"] = int(m.group(3))
                        data["read2"]["rr"] = int(m.group(4))

        return data

    def chart_align_strand(self):
        """
        Chart _strand_table.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Mapping Strand Distribution chart
        """

        pd1 = {}
        pd2 = {}
        for s_name, dd in self.mdata["align_strand"].items():
            if len(dd["read1"]) > 0:
                pd1[s_name] = dd["read1"]
            if len(dd["read2"]) > 0:
                pd2[s_name] = dd["read2"]

        pheader = {
            "ff": {"color": "#F53855", "name": "ff: Watson-Aligned, Watson-Bisulfite Conversion"},
            "fr": {"color": "#E37B40", "name": "fr: Watson-Aligned, Crick-Bisulfite Conversion"},
            "rf": {"color": "#46B29D", "name": "rf: Crick-Aligned, Watson-Bisulfite Conversion"},
            "rr": {"color": "#324D5C", "name": "rr: Crick-Aligned, Crick-Bisulfite Conversion"},
        }
        pconfig = {
            "id": "biscuit_strands",
            "title": "BISCUIT: Mapping Strand Distribution",
            "ylab": "Number of Reads",
            "cpswitch_counts_label": "# Reads",
            "data_labels": [{"name": "Read 1"}, {"name": "Read 2"}],
        }

        # TODO: When PBAT mode is implemented, add comment in help text about
        #       how to interpret PBAT mode results
        if max(len(pd1), len(pd2)) > 0:
            self.add_section(
                name="Mapping Strand Distribution",
                anchor="biscuit-strands",
                description="For primary alignments, shows the number of reads mapped to each strand.",
                helptext="""
                    Most bisulfite libraries typically map Read 1 to the parent
                    strand (`ff`, `rr`) and Read 2 to the daughter / synthesized
                    strand (`fr`, `rf`).

                    Note that PBAT and many single-cell / low input
                    libraries may not follow this assumption.
                """,
                plot=bargraph.plot([pd1, pd2], [pheader, pheader], pconfig),
            )

    ########################################
    ####       Insert Size Report       ####
    ########################################
    @staticmethod
    def parse_logs_align_isize(f, fn):
        """
        Parse _isize_table.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of insert size data
        """
        file_data = f.splitlines()[2:]

        # Handle missing data
        if len(file_data) == 0:
            log.debug(f"No data available in {fn}. Will not fill corresponding entries.")
            return {"no_data_available": 1}

        data = {"percent": {}, "readcnt": {}}
        for line in file_data:
            fields = line.split("\t")
            data["percent"][int(fields[0])] = 100.0 * float(fields[1])
            data["readcnt"][int(fields[0])] = float(fields[2])

        return data

    def chart_align_isize(self):
        """
        Chart _isize_table.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Insert Size Distribution chart
        """

        pd_p = {}
        pd_r = {}
        for s_name, dd in self.mdata["align_isize"].items():
            if "no_data_available" not in dd.keys():
                pd_p[s_name] = dd["percent"]
                pd_r[s_name] = {ins: cnt * config.read_count_multiplier for ins, cnt in dd["readcnt"].items()}

        pconfig = {
            "id": "biscuit_isize",
            "title": "BISCUIT: Insert Size Distribution",
            "ymin": 0,
            "xmin": 0,
            "yLabelFormat": "{value}",
            "smooth_points": 1000,  # limit number of points / smooth data
            "tt_label": "<strong>IS{point.x}:</strong> {point.y:.2f}",
            "xlab": "Insert Size",
            "ylab": "% of Mapped Reads",
            "data_labels": [
                {"name": "Percent of Reads", "ylab": "% of Mapped Reads"},
                {
                    "name": f"{config.read_count_desc.capitalize()} of Reads",
                    "ylab": f"{config.read_count_desc.capitalize()} of Mapped Reads",
                },
            ],
        }

        self.add_section(
            name="Insert Size Distribution",
            anchor="biscuit-isize",
            description="Shows the distribution of insert sizes.",
            helptext="""
                Insert size is defined as:

                ```
                (right-most coordinate of reverse-mate read) - (left-most coordinate of forward-mate read)
                ```

                Insert sizes are calculated for reads with a _"mapped in
                proper pair"_ `samtools` flag, and `MAPQ >= 40`.
            """,
            plot=linegraph.plot([pd_p, pd_r], pconfig),
        )

    ########################################
    ####        Duplicate Report        ####
    ########################################
    @staticmethod
    def parse_logs_dup_report(f, fn):
        """
        Parses _dup_report.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of duplicate fractions
        """
        patterns = [
            (r"Number of duplicate reads:\s+(\d+)", r"Number of reads:\s+(\d+)", "all"),
            (r"Number of duplicate q40-reads:\s+(\d+)", r"Number of q40-reads:\s+(\d+)", "q40"),
        ]

        data = {}
        for pat_dup, pat_tot, key in patterns:
            m1 = re.search(pat_dup, f, re.MULTILINE)
            m2 = re.search(pat_tot, f, re.MULTILINE)
            if m1 is not None and m2 is not None:
                data[key] = 100.0 * float(m1.group(1)) / float(m2.group(1))
            else:
                data[key] = -1

        return data

    def chart_dup_report(self):
        """
        Charts _dup_report.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Duplicate Rates chart
        """

        pd1 = {}  # Overall duplicate rate
        pd2 = {}  # MAPQ>=40 duplicate rate
        for s_name, dd in self.mdata["dup_report"].items():
            if "all" in dd and dd["all"] != -1:
                pd1[s_name] = {"dup_rate": dd["all"]}
            if "q40" in dd and dd["q40"] != -1:
                pd2[s_name] = {"dup_rate": dd["q40"]}

        pheader = {
            "dup_rate": {"color": "#a50f15", "name": "Duplicate Rate"},
        }
        pconfig = {
            "id": "biscuit_dup_report",
            "cpswitch": False,
            "cpswitch_c_active": False,
            "title": "BISCUIT: Percentage of Duplicate Reads",
            "data_labels": [{"name": "Overall Duplicate Rate"}, {"name": "MAPQ>=40 Duplicate Rate"}],
            "ylab": "Duplicate Rate [%]",
            "ymin": 0,
            "ymax": 100,
            "yCeiling": 110,
            "use_legend": False,
            "tt_decimals": 1,
            "tt_suffix": "%",
            "tt_percentages": False,
        }

        if len(pd1) > 0:
            self.add_section(
                name="Duplicate Rates",
                anchor="biscuit-dup-report",
                description="Shows the percentage of total reads that are duplicates.",
                helptext="""
                    `MAPQ >= 40` shows the duplicate rate for just the reads reads
                    with a mapping quality score of `MAPQ >= 40`.
                """,
                plot=bargraph.plot([pd1, pd2], [pheader, pheader], pconfig),
            )

    ########################################
    ####      Depths and Uniformity     ####
    ########################################
    @staticmethod
    def parse_logs_qc_cv(f, fn):
        """
        Parses _cv_table.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of depth uniformity measures
        """

        data = {}
        targets = [
            "all_base",
            "all_cpg",
            "q40_base",
            "q40_cpg",
            "all_base_botgc",
            "all_cpg_botgc",
            "q40_base_botgc",
            "q40_cpg_botgc",
            "all_base_topgc",
            "all_cpg_topgc",
            "q40_base_topgc",
            "q40_cpg_topgc",
        ]
        for t in targets:
            m = re.search(rf"{t}\t([\d\.]+)\t([\d\.]+)\t([\d\.]+)", f, re.MULTILINE)
            if m is not None:
                data[t] = {"mu": float(m.group(1)), "sigma": float(m.group(2)), "cv": float(m.group(3))}
            else:
                data[t] = {"mu": -1, "sigma": -1, "cv": -1}

        return data

    def chart_qc_cv(self):
        """
        Charts _cv_table.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Sequencing Depth - Whole Genome chart
        """

        cats = [
            ("all_base", "a_b"),
            ("q40_base", "q_b"),
            ("all_base_botgc", "a_b_b"),
            ("q40_base_botgc", "q_b_b"),
            ("all_base_topgc", "a_b_t"),
            ("q40_base_topgc", "q_b_t"),
            ("all_cpg", "a_c"),
            ("q40_cpg", "q_c"),
            ("all_cpg_botgc", "a_c_b"),
            ("q40_cpg_botgc", "q_c_b"),
            ("all_cpg_topgc", "a_c_t"),
            ("q40_cpg_topgc", "q_c_t"),
        ]

        pd = dict()
        for s_name, dd in self.mdata["qc_cv"].items():
            data = dict()
            for cat, key in cats:
                if cat in dd:
                    if dd[cat]["mu"] != -1:
                        data["mu_" + key] = dd[cat]["mu"]
                        data["cv_" + key] = dd[cat]["cv"]
            if len(data) > 0:
                pd[s_name] = data

        shared_mean = {"min": 0, "format": "{:,3f}", "minRange": 10}
        shared_cofv = {"min": 0, "format": "{:,3f}", "minRange": 50}

        pheader = {
            "mu_a_b": dict(
                shared_mean, **{"title": "All Genome Mean", "description": "Mean Sequencing Depth for All Reads"}
            ),
            "mu_q_b": dict(
                shared_mean, **{"title": "Q40 Genome Mean", "description": "Mean Sequencing Depth for Q40 Reads"}
            ),
            "mu_a_b_b": dict(
                shared_mean,
                **{
                    "title": "Low GC All Gen. Mean",
                    "description": "Mean Sequencing Depth for All Reads in Low GC-Content Regions",
                },
            ),
            "mu_q_b_b": dict(
                shared_mean,
                **{
                    "title": "Low GC Q40 Gen. Mean",
                    "description": "Mean Sequencing Depth for Q40 Reads in Low GC-Content Regions",
                },
            ),
            "mu_a_b_t": dict(
                shared_mean,
                **{
                    "title": "High GC All Gen. Mean",
                    "description": "Mean Sequencing Depth for All Reads in High GC-Content Regions",
                },
            ),
            "mu_q_b_t": dict(
                shared_mean,
                **{
                    "title": "High GC Q40 Gen. Mean",
                    "description": "Mean Sequencing Depth for Q40 Reads in High GC-Content Regions",
                },
            ),
            "cv_a_b": dict(
                shared_cofv, **{"title": "All Genome CoV", "description": "Sequencing Depth CoV for All Reads"}
            ),
            "cv_q_b": dict(
                shared_cofv, **{"title": "Q40 Genome CoV", "description": "Sequencing Depth CoV for Q40 Reads"}
            ),
            "cv_a_b_b": dict(
                shared_cofv,
                **{
                    "title": "Low GC All Gen. CoV",
                    "description": "Sequencing Depth CoV for All Reads in Low GC-Content Regions",
                },
            ),
            "cv_q_b_b": dict(
                shared_cofv,
                **{
                    "title": "Low GC Q40 Gen. CoV",
                    "description": "Sequencing Depth CoV for Q40 Reads in Low GC-Content Regions",
                },
            ),
            "cv_a_b_t": dict(
                shared_cofv,
                **{
                    "title": "High GC All Gen. CoV",
                    "description": "Sequencing Depth CoV for All Reads in High GC-Content Regions",
                },
            ),
            "cv_q_b_t": dict(
                shared_cofv,
                **{
                    "title": "High GC Q40 Gen. CoV",
                    "description": "Sequencing Depth CoV for Q40 Reads in High GC-Content Regions",
                },
            ),
            "mu_a_c": dict(
                shared_mean, **{"title": "All CpGs Mean", "description": "Mean Sequencing Depth for All CpGs"}
            ),
            "mu_q_c": dict(
                shared_mean, **{"title": "Q40 CpGs Mean", "description": "Mean Sequencing Depth for Q40 CpGs"}
            ),
            "mu_a_c_b": dict(
                shared_mean,
                **{
                    "title": "Low GC All CpGs Mean",
                    "description": "Mean Sequencing Depth for All CpGs in Low GC-Content Regions",
                },
            ),
            "mu_q_c_b": dict(
                shared_mean,
                **{
                    "title": "Low GC Q40 CpGs Mean",
                    "description": "Mean Sequencing Depth for Q40 CpGs in Low GC-Content Regions",
                },
            ),
            "mu_a_c_t": dict(
                shared_mean,
                **{
                    "title": "High GC All CpGs Mean",
                    "description": "Mean Sequencing Depth for All CpGs in High GC-Content Regions",
                },
            ),
            "mu_q_c_t": dict(
                shared_mean,
                **{
                    "title": "High GC Q40 CpGs Mean",
                    "description": "Mean Sequencing Depth for Q40 CpGs in High GC-Content Regions",
                },
            ),
            "cv_a_c": dict(
                shared_cofv, **{"title": "All CpGs CoV", "description": "Sequencing Depth CoV for All CpGs"}
            ),
            "cv_q_c": dict(
                shared_cofv, **{"title": "Q40 CpGs CoV", "description": "Sequencing Depth CoV for Q40 CpGs"}
            ),
            "cv_a_c_b": dict(
                shared_cofv,
                **{
                    "title": "Low GC All CpGs CoV",
                    "description": "Sequencing Depth CoV for All CpGs in Low GC-Content Regions",
                },
            ),
            "cv_q_c_b": dict(
                shared_cofv,
                **{
                    "title": "Low GC Q40 CpGs CoV",
                    "description": "Sequencing Depth CoV for Q40 CpGs in Low GC-Content Regions",
                },
            ),
            "cv_a_c_t": dict(
                shared_cofv,
                **{
                    "title": "High GC All CpGs CoV",
                    "description": "Sequencing Depth CoV for All CpGs in High GC-Content Regions",
                },
            ),
            "cv_q_c_t": dict(
                shared_cofv,
                **{
                    "title": "High GC Q40 CpGs CoV",
                    "description": "Sequencing Depth CoV for Q40 CpGs in High GC-Content Regions",
                },
            ),
        }
        pconfig = {
            "id": "biscuit_seq_depth",
            "table_title": "BISCUIT: Sequencing Depth",
            "sortRows": False,
        }

        if len(pd) > 0:
            self.add_section(
                name="Sequencing Depth Statistics",
                anchor="biscuit-seq-depth",
                description="""
                    Shows the sequence depth mean and uniformity measured by the Coefficient of Variation
                    (`CoV`, defined as `stddev/mean`).
                """,
                helptext="""
                    The plot shows coverage across different selections:

                    * _Genome_ (Gen.) - Statistics for all bases across the entire genome
                    * _CpGs_ - Statistics for CpGs
                    * _All_ - Statistics for any mapped bases/CpGs
                    * _Q40_ - Statistics only those bases/CpGs with mapping quality `MAPQ >= 40`
                    * _High GC_ - Bases / CpGs that overlap with the top 10% of 100bp windows for GC-content
                    * _Low GC_ - Bases / CpGs that overlap with the bottom 10% of 100bp windows for GC-content

                """,
                plot=beeswarm.plot(pd, pheader, pconfig),
            )

    ########################################
    #### Base Coverage and CpG Coverage ####
    ########################################
    @staticmethod
    def parse_logs_covdist_all_base(f, fn):
        """
        Parses _covdist_all_base_botgc_table.txt
               _covdist_all_base_table.txt
               _covdist_all_base_topgc_table.txt
               _covdist_all_cpg_botgc_table.txt
               _covdist_all_cpg_table.txt
               _covdist_all_cpg_topgc_table.txt
               _covdist_q40_base_botgc_table.txt
               _covdist_q40_base_table.txt
               _covdist_q40_base_topgc_table.txt
               _covdist_q40_cpg_botgc_table.txt
               _covdist_q40_cpg_table.txt
               _covdist_q40_cpg_topgc_table.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of coverage distributions up to 30X data
        """
        file_data = f.splitlines()[2:]

        # Handle missing data
        if len(file_data) == 0:
            log.debug(f"No data available in {fn}. Will not fill corresponding entries.")
            return dict(zip([i for i in range(31)], [-1 for _ in range(31)]))

        dd = {}
        for line in file_data:
            fields = line.split()
            dd[int(float(fields[0]))] = int(float(fields[1]))

        covs = sorted([k for k in dd])[:31]
        _ccov_cnt = sum(dd.values())

        ccov_cnts = []
        for cov in covs:
            ccov_cnts.append(_ccov_cnt / 1000000.0)
            _ccov_cnt -= dd[cov]

        return dict(zip(covs, ccov_cnts))

    def parse_logs_covdist_all_base_botgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_all_base_topgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_q40_base(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_q40_base_botgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_q40_base_topgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_all_cpg(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_all_cpg_botgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_all_cpg_topgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_q40_cpg(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_q40_cpg_botgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def parse_logs_covdist_q40_cpg_topgc(self, f, fn):
        """Handled by parse_logs_covdist_all_base()"""
        return self.parse_logs_covdist_all_base(f, fn)

    def chart_covdist_all_base(self):
        """
        Charts _covdist_all_base_botgc_table.txt
               _covdist_all_base_table.txt
               _covdist_all_base_topgc_table.txt
               _covdist_all_cpg_botgc_table.txt
               _covdist_all_cpg_table.txt
               _covdist_all_cpg_topgc_table.txt
               _covdist_q40_base_botgc_table.txt
               _covdist_q40_base_table.txt
               _covdist_q40_base_topgc_table.txt
               _covdist_q40_cpg_botgc_table.txt
               _covdist_q40_cpg_table.txt
               _covdist_q40_cpg_topgc_table.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Cumulative Coverage chart
        """

        pd = [
            self.mdata["covdist_all_base"],
            self.mdata["covdist_q40_base"],
            self.mdata["covdist_all_cpg"],
            self.mdata["covdist_q40_cpg"],
            self.mdata["covdist_all_base_botgc"],
            self.mdata["covdist_q40_base_botgc"],
            self.mdata["covdist_all_cpg_botgc"],
            self.mdata["covdist_q40_cpg_botgc"],
            self.mdata["covdist_all_base_topgc"],
            self.mdata["covdist_q40_base_topgc"],
            self.mdata["covdist_all_cpg_topgc"],
            self.mdata["covdist_q40_cpg_topgc"],
        ]

        pconfig = {
            "id": "biscuit_cumulative",
            "title": "BISCUIT: Cumulative Coverage",
            "ymin": 0,
            "tt_label": "<strong>{point.x}X:</strong> {point.y:.2f}M",
            "xlab": "Coverage",
            "ylab": "Millions of Bases",
            "data_labels": [
                {"name": "All Bases", "ylab": "Millions of Bases"},
                {"name": "Q40 Bases", "ylab": "Millions of Bases"},
                {"name": "All CpGs", "ylab": "Millions of CpGs"},
                {"name": "Q40 CpGs", "ylab": "Millions of CpGs"},
                {"name": "Low GC All Bases", "ylab": "Millions of Bases"},
                {"name": "Low GC Q40 Bases", "ylab": "Millions of Bases"},
                {"name": "Low GC All CpGs", "ylab": "Millions of CpGs"},
                {"name": "Low GC Q40 CpGs", "ylab": "Millions of CpGs"},
                {"name": "High GC All Bases", "ylab": "Millions of Bases"},
                {"name": "High GC Q40 Bases", "ylab": "Millions of Bases"},
                {"name": "High GC All CpGs", "ylab": "Millions of CpGs"},
                {"name": "High GC Q40 CpGs", "ylab": "Millions of CpGs"},
            ],
        }

        self.add_section(
            name="Cumulative Coverage",
            anchor="biscuit-cumulative-coverage",
            description="Shows the number of bases or CpGs covered by a given number of reads.",
            helptext="""
                * _All_ - Coverage for any mapped reads
                * _Q40_ - Coverage for reads with mapping quality `MAPQ >= 40`
                * _High GC_ - Coverage for reads that overlap with the top 10% of 100bp windows for GC-content
                * _Low GC_ - Coverage for reads that overlap with the bottom 10% of 100bp windows for GC-content
            """,
            plot=linegraph.plot(pd, pconfig),
        )

    def chart_covdist_all_base_botgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_all_base_topgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_q40_base(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_q40_base_botgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_q40_base_topgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_all_cpg(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_all_cpg_botgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_all_cpg_topgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_q40_cpg(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_q40_cpg_botgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    def chart_covdist_q40_cpg_topgc(self):
        """Handled by chart_covdist_all_base()"""
        pass

    ########################################
    ####          CpG Retention         ####
    ########################################
    def parse_logs_cpg_retention_readpos(self, f, fn):
        """
        Parses _CpGRetentionByReadPos.txt
               _CpHRetentionByReadPos.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of fraction of retained cytosines for reads 1 and 2
            in either a CpH or CpG context
        """
        file_data = f.splitlines()[2:]

        # Handle missing data
        if len(file_data) == 0:
            log.debug(f"No data available in {fn}. Will not fill corresponding entries.")
            return {"no_data_available": 1}

        r1 = {"C": {}, "R": {}}
        r2 = {"C": {}, "R": {}}
        for line in file_data:
            fields = line.strip().split("\t")

            if fields[0] not in ["1", "2"] or fields[2] not in ["C", "R"]:
                return {}
            if fields[0] == "1":
                r1[fields[2]][int(fields[1])] = int(fields[3])
            elif fields[0] == "2":
                r2[fields[2]][int(fields[1])] = int(fields[3])

        r1rate = dict()
        for k in sorted(r1["C"].keys()):
            if k in r1["R"]:
                r1rate[k] = 100.0 * float(r1["R"][k]) / (r1["R"][k] + r1["C"][k])

        r2rate = dict()
        for k in sorted(r2["C"].keys()):
            if k in r2["R"]:
                r2rate[k] = 100.0 * float(r2["R"][k]) / (r2["R"][k] + r2["C"][k])

        return {"1": r1rate, "2": r2rate}

    def chart_cpg_retention_readpos(self):
        """
        Charts _CpGRetentionByReadPos.txt
               _CpHRetentionByReadPos.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Retenion vs. Base Position in Read chart
        """

        pd = [
            dict(
                [
                    (s_name, dd["1"])
                    for s_name, dd in self.mdata["cpg_retention_readpos"].items()
                    if "no_data_available" not in dd.keys()
                ]
            ),
            dict(
                [
                    (s_name, dd["2"])
                    for s_name, dd in self.mdata["cpg_retention_readpos"].items()
                    if "no_data_available" not in dd.keys()
                ]
            ),
            dict(
                [
                    (s_name, dd["1"])
                    for s_name, dd in self.mdata["cph_retention_readpos"].items()
                    if "no_data_available" not in dd.keys()
                ]
            ),
            dict(
                [
                    (s_name, dd["2"])
                    for s_name, dd in self.mdata["cph_retention_readpos"].items()
                    if "no_data_available" not in dd.keys()
                ]
            ),
        ]

        pconfig = {
            "id": "biscuit_retention_cytosine",
            "title": "BISCUIT: Retention vs. Base Position in Read",
            "xlab": "Position in Read",
            "ylab": "CpG Retention Rate (%)",
            "ymin": 0,
            "ymax": 100,
            "yMinRange": 0,
            "yFloor": 0,
            "tt_label": "<strong>Position {point.x}:</strong> {point.y:.2f}%",
            "data_labels": [
                {"name": "CpG Read 1", "ylab": "CpG Retention Rate (%)"},
                {"name": "CpG Read 2", "ylab": "CpG Retention Rate (%)"},
                {"name": "CpH Read 1", "ylab": "CpH Retention Rate (%)"},
                {"name": "CpH Read 2", "ylab": "CpH Retention Rate (%)"},
            ],
        }

        self.add_section(
            name="Retention vs. Base Position in Read",
            anchor="biscuit-retention-cytosine",
            description="Distribution of cytosine retention rates across base positions in the read (a.k.a. _M-bias_ plot).",
            plot=linegraph.plot(pd, pconfig),
        )

    def parse_logs_cph_retention_readpos(self, f, fn):
        """Handled by parse_logs_cpg_retention_readpos()"""
        return self.parse_logs_cpg_retention_readpos(f, fn)

    def chart_cph_retention_readpos(self):
        """Handled by chart_cpg_retention_readpos()"""
        pass

    def parse_logs_read_avg_retention_rate(self, f, fn):
        """
        Parses _totalReadConversionRate.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of read averaged fraction of retainied cytosines by context
        """

        file_data = f.splitlines()[2:]

        # Handle missing data
        if len(file_data) == 0:
            log.debug(f"No data available in {fn}. Will not fill corresponding entries.")
            return {"no_data_available": 1}

        data = {}
        for line in file_data:
            fields = line.split("\t")
            # Skip rows that have NaNs as something went wrong in processing
            if "nan" in fields:
                log.debug(f"Found NaN in {fn}. Skipping.")
                continue

            # BISCUIT returns -1 if insufficient data. Only fill fields with value >= 0.
            if float(fields[0]) >= 0:
                data["rca"] = 100.0 * float(fields[0])
            if float(fields[1]) >= 0:
                data["rcc"] = 100.0 * float(fields[1])
            if float(fields[2]) >= 0:
                data["rcg"] = 100.0 * float(fields[2])
            if float(fields[3]) >= 0:
                data["rct"] = 100.0 * float(fields[3])

        return data

    def chart_read_avg_retention_rate(self):
        """
        Charts _totalReadConversionRate.txt
               _totalBaseConversionRate.txt
        Inputs:
            No inputs
        Returns:
            No returns, generates Cytosine Retention chart
        """

        pdata_byread = {}
        for s_name, dd in self.mdata["read_avg_retention_rate"].items():
            if "no_data_available" not in dd.keys():
                pdata_byread[s_name] = dd

        pdata_bybase = {}
        for s_name, dd in self.mdata["base_avg_retention_rate"].items():
            if "no_data_available" not in dd.keys():
                pdata_bybase[s_name] = dd

        pheader_byread = {
            "rca": {"color": "#D81B60", "name": "CpA Retention"},
            "rcc": {"color": "#1E88E5", "name": "CpC Retention"},
            "rcg": {"color": "#A0522D", "name": "CpG Retention"},
            "rct": {"color": "#004D40", "name": "CpT Retention"},
        }
        pheader_bybase = {
            "bca": {"color": "#D81B60", "name": "CpA Retention"},
            "bcc": {"color": "#1E88E5", "name": "CpC Retention"},
            "bcg": {"color": "#A0522D", "name": "CpG Retention"},
            "bct": {"color": "#004D40", "name": "CpT Retention"},
        }
        pconfig = {
            "id": "biscuit_retention",
            "cpswitch": False,
            "cpswitch_c_active": False,
            "title": "BISCUIT: Cytosine Retention",
            "data_labels": [{"name": "Read-averaged Retention"}, {"name": "Base-averaged Retention"}],
            "ylab": "Percent Retained",
            "ymin": 0,
            "ymax": 100,
            "yCeiling": 110,
            "stacking": None,
            "tt_decimals": 1,
            "tt_suffix": "%",
        }

        self.add_section(
            name="Cytosine Retention",
            anchor="biscuit-retention",
            description="Shows the cytosine retention rate for different contexts.",
            helptext="""
                The cytosine retention rate is calculated as `1 - (cytosine conversion rate)`.

                Assuming complete, but not over, bisulfite conversion, the cytosine retention rate
                is the average cytosine modification (including 5mC, 5hmC, etc) rate.

                Note, if a sample is missing from the Base-averaged Retention table,
                there wasn't sufficient data to plot that sample.
            """,
            plot=bargraph.plot([pdata_byread, pdata_bybase], [pheader_byread, pheader_bybase], pconfig),
        )

    def parse_logs_base_avg_retention_rate(self, f, fn):
        """
        Parses _totalBaseConversionRate.txt
        Inputs:
            f - current matched file
            fn - filename
        Returns:
            data - dictionary of base averaged fraction of retainied cytosines by context
        """

        file_data = f.splitlines()[2:]

        # Handle missing data
        if len(file_data) == 0:
            log.debug(f"No data available in {fn}. Will not fill corresponding entries.")
            return {"no_data_available": 1}

        data = {}
        for line in file_data:
            fields = line.split("\t")
            # Skip rows that have NaNs as something went wrong in processing
            if "nan" in fields:
                log.debug(f"Found NaN in {fn}. Skipping.")
                continue

            # BISCUIT returns -1 if insufficient data. Only fill fields with value >= 0.
            if float(fields[0]) >= 0:
                data["bca"] = 100.0 * float(fields[0])
            if float(fields[1]) >= 0:
                data["bcc"] = 100.0 * float(fields[1])
            if float(fields[2]) >= 0:
                data["bcg"] = 100.0 * float(fields[2])
            if float(fields[3]) >= 0:
                data["bct"] = 100.0 * float(fields[3])

        return data

    def chart_base_avg_retention_rate(self):
        """Handled by chart_read_avg_retention_rate()"""
        pass



import copy
import dataclasses
import logging
from typing import Dict, List, Union

import plotly.graph_objects as go

from multiqc.plots.plotly import determine_barplot_height
from multiqc.plots.plotly.plot import Plot, PlotType, BaseDataset
from multiqc.utils import util_functions

logger = logging.getLogger(__name__)


# Type of single box (matching one sample)
BoxT = List[Union[int, float]]


def plot(list_of_data_by_sample: List[Dict[str, BoxT]], pconfig: Dict) -> str:
    """
    Build and add the plot data to the report, return an HTML wrapper.
    :param list_of_data_by_sample: each dataset is a dict mapping samples to either:
        * list of data points,
        * list of statistics (e.g. {min, max, median, mean, std, q1, q3 etc.})
    :param pconfig: Plot configuration dictionary
    :return: HTML with JS, ready to be inserted into the page
    """
    p = BoxPlot(
        pconfig,
        list_of_data_by_sample,
        max_n_samples=max(len(d) for d in list_of_data_by_sample),
    )

    from multiqc.utils import report

    return p.add_to_report(report)


class BoxPlot(Plot):
    @dataclasses.dataclass
    class Dataset(BaseDataset):
        data: List[BoxT]
        samples: List[str]

        @staticmethod
        def create(
            dataset: BaseDataset,
            data_by_sample: Dict[str, BoxT],
        ) -> "BoxPlot.Dataset":
            dataset = BoxPlot.Dataset(
                **dataset.__dict__,
                data=list(data_by_sample.values()),
                samples=list(data_by_sample.keys()),
            )
            # Need to reverse samples as the box plot will show them reversed
            dataset.samples = list(reversed(dataset.samples))
            dataset.data = list(reversed(dataset.data))

            dataset.trace_params.update(
                boxpoints="outliers",
                jitter=0.5,
                orientation="h",
                marker=dict(
                    color="#4899e8",  # just use blue to indicate interactivity
                ),
                # to remove the redundant sample name before "median" in the unified hover box
                hoverinfo="x",
            )
            dataset.layout["yaxis"]["title"] = None
            return dataset

        def create_figure(
            self,
            layout: go.Layout,
            is_log=False,
            is_pct=False,
        ) -> go.Figure:
            """
            Create a Plotly figure for a dataset
            """
            fig = go.Figure(layout=layout)

            for sname, values in zip(self.samples, self.data):
                params = copy.deepcopy(self.trace_params)
                fig.add_trace(
                    go.Box(
                        x=values,
                        name=sname,
                        **params,
                    ),
                )
            return fig

    def __init__(
        self,
        pconfig: Dict,
        list_of_data_by_sample: List[Dict[str, BoxT]],
        max_n_samples: int,
    ):
        super().__init__(PlotType.BOX, pconfig, n_datasets=len(list_of_data_by_sample))

        self.datasets: List[BoxPlot.Dataset] = [
            BoxPlot.Dataset.create(ds, data_by_sample)
            for ds, data_by_sample in zip(self.datasets, list_of_data_by_sample)
        ]

        height = determine_barplot_height(max_n_samples)

        self.layout.update(
            height=height,
            showlegend=False,
            boxgroupgap=0.1,
            boxgap=0.2,
            colorway=[],  # no need to color code
            yaxis=dict(
                automargin=True,  # to make sure there is enough space for ticks labels
                categoryorder="trace",  # keep sample order
                hoverformat=self.layout.xaxis.hoverformat,
                ticksuffix=self.layout.xaxis.ticksuffix,
                # Prevent JavaScript from automatically parsing categorical values as numbers:
                type="category",
            ),
            xaxis=dict(
                title=dict(text=self.layout.yaxis.title.text),
                hoverformat=self.layout.yaxis.hoverformat,
                ticksuffix=self.layout.yaxis.ticksuffix,
            ),
            hovermode="y",
            hoverlabel=dict(
                bgcolor="rgba(255, 255, 255, 0.8)",
                font=dict(color="black"),
            ),
        )

    def save_data_file(self, dataset: Dataset) -> None:
        vals_by_sample = {}
        for sample, values in zip(dataset.samples, dataset.data):
            vals_by_sample[sample] = values
        util_functions.write_data_file(vals_by_sample, dataset.uid)



""" MultiQC module to parse output from mtnucratio """


import json
import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """mtnucratio module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="mtnucratio",
            anchor="mtnucratio",
            href="http://www.github.com/apeltzer/MTNucRatioCalculator",
            info="is a tool to compute mt/nuc ratios for NGS datasets.",
            doi="10.1186/s13059-016-0918-z",
        )

        # Find and load any MTNUCRATIO reports
        self.mtnuc_data = dict()

        for f in self.find_log_files("mtnucratio", filehandles=True):
            self.parseJSON(f)

        # Filter to strip out ignored sample names
        self.mtnuc_data = self.ignore_samples(self.mtnuc_data)

        if len(self.mtnuc_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.mtnuc_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.mtnuc_data, "multiqc_mtnucratio")

        # Basic Stats Table
        self.mtnucratio_general_stats_table()

    # Parse our nice little JSON file
    def parseJSON(self, f):
        """Parse the JSON output from mtnucratio and save the summary statistics"""
        try:
            parsed_json = json.load(f["f"])
            if "metrics" not in parsed_json and "metadata" not in parsed_json:
                log.warning(f"No MTNUCRATIO JSON: '{f['fn']}'")
                return None
        except json.JSONDecodeError as e:
            log.warning(f"Could not parse mtnucratio JSON: '{f['fn']}'")
            log.debug(e)
            return None

        # Get sample name from JSON first
        s_name = self.clean_s_name(parsed_json["metadata"]["sample_name"], f)
        self.add_data_source(f, s_name)

        metrics_dict = parsed_json["metrics"]

        # Add all in the main data_table
        self.mtnuc_data[s_name] = metrics_dict

        # Add version info
        version = parsed_json["metadata"]["version"]
        self.add_software_version(version, s_name)

    def mtnucratio_general_stats_table(self):
        """Take the parsed stats from the mtnucratio report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "mt_cov_avg": {
                "title": "MT genome coverage",
                "description": "Average coverage (X) on mitochondrial genome.",
                "min": 0,
                "scale": "OrRd",
                "suffix": " X",
                "hidden": True,
            },
            "nuc_cov_avg": {
                "title": "Genome coverage",
                "description": "Average coverage (X) on nuclear genome.",
                "min": 0,
                "scale": "GnBu",
                "suffix": " X",
                "hidden": True,
            },
            "mt_nuc_ratio": {
                "title": "MT to Nuclear Ratio",
                "description": "Mitochondrial to nuclear reads ratio (MTNUC)",
                "min": 0,
                "max": 100,
                "scale": "RdYlGn-rev",
            },
            "nucreads": {
                "title": f"{config.read_count_prefix} Genome reads",
                "description": f"Reads on the nuclear genome ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "BuPu",
                "hidden": True,
            },
            "mtreads": {
                "title": f"{config.read_count_prefix} MT genome reads",
                "description": f"Reads on the mitochondrial genome ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "OrRd",
                "hidden": True,
            },
        }

        self.general_stats_addcols(self.mtnuc_data, headers)



""" MultiQC module to parse output from eigenstrat_snp_coverage """


import json
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """eigenstratdatabasetools module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="eigenstratdatabasetools",
            anchor="eigenstrat",
            href="https://github.com/TCLamnidis/EigenStratDatabaseTools",
            info="A set of tools to compare and manipulate the contents of EingenStrat databases, and to calculate SNP coverage statistics in such databases.",
            # No publication / DOI // doi=
        )

        # Find and load any DeDup reports
        self.snp_cov_data = dict()

        # Find and load JSON file
        for f in self.find_log_files("eigenstratdatabasetools", filehandles=True):
            self.parse_data(f)

        # Filter samples
        self.snp_cov_data = self.ignore_samples(self.snp_cov_data)

        # Return if no samples found
        if len(self.snp_cov_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.snp_cov_data)} reports")

        # Save data output file
        self.write_data_file(self.snp_cov_data, "multiqc_snp_cov_metrics")

        # Add to General Statistics
        self.addSummaryMetrics()

    def parse_data(self, f):
        try:
            data = json.load(f["f"])
        except Exception as e:
            log.debug(e)
            log.warning(f"Could not parse eigenstrat_snp_coverage JSON: '{f['fn']}'")
            return

        # Parse JSON data to a dict
        version = None
        for s_name in data:
            if s_name == "Metadata":
                version = data["Metadata"]["version"]
                continue

            s_clean = self.clean_s_name(s_name, f)
            if s_clean in self.snp_cov_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_clean}")

            if version is not None:
                self.add_software_version(version, s_clean)

            self.add_data_source(f, s_clean)
            self.snp_cov_data[s_clean] = dict()

            for k, v in data[s_name].items():
                try:
                    # number of covered and total SNPs are integer values
                    self.snp_cov_data[s_clean][k] = int(v)
                except ValueError:
                    self.snp_cov_data[s_clean][k] = v

    def addSummaryMetrics(self):
        """Take the parsed stats from eigenstrat_snp_coverage and add it to the main plot"""

        headers = {
            "Covered_Snps": {
                "title": "Covered SNPs",
                "description": "The number of SNPs for which a genotype has been called.",
                "scale": "PuBuGn",
                "format": "{:,.0f}",
                "shared_key": "snp_call",
            },
            "Total_Snps": {
                "title": "Total SNPs",
                "description": "The total number of SNPs in the genotype dataset.",
                "scale": "PuBuGn",
                "format": "{:,.0f}",
                "hidden": True,
                "shared_key": "snp_call",
            },
        }

        self.general_stats_addcols(self.snp_cov_data, headers)



# -*- coding: utf-8 -*-

""" MultiQC module to parse output files from iVar """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import heatmap

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="iVar",
            anchor="iVar",
            href="https://github.com/andersen-lab/ivar",
            info="is a computational package that contains functions broadly useful for viral amplicon-based sequencing.",
            doi="10.1101/383513",
        )

        # Find and load iVar trim results
        self.ivar_data = dict()
        self.ivar_primers = dict()
        for f in self.find_log_files("ivar/trim", filehandles=True):
            self.parse_ivar(f)

        # Filter to strip out ignored sample names
        self.ivar_data = self.ignore_samples(self.ivar_data)
        self.ivar_primers = self.ignore_samples(self.ivar_primers)

        # Stop if no files are found
        num_samples = max(len(self.ivar_data), len(self.ivar_primers))
        if num_samples == 0:
            raise ModuleNoSamplesFound

        # Log number of reports
        log.info(f"Found {num_samples} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed data to a file
        self.write_data_file(self.ivar_data, "multiqc_ivar_summary")
        self.write_data_file(self.ivar_primers, "multiqc_ivar_primers")

        # Basic Stats Table
        self.ivar_general_stats_table()

        # Heatmap info
        self.primer_heatmap()

    # Parse an ivar report
    def parse_ivar(self, f):
        """Parse iVar log file, either v1.1_beta or v1.2"""
        count_regexes = {
            "mapped_reads": re.compile(r"(?:Found\s)(\d+)(?:\smapped)"),
            "trimmed_reads": re.compile(r"(?:Trimmed primers from )(?:\d+\.\d+\% \()?(\d+)"),
            "reads_outside_primer_region": re.compile(
                r"^(?:\d+\.\d+\% \()?(\d+)(?:\))?(?:.*[of]?)reads\s(?:that\s)?started"
            ),
            "reads_too_short_after_trimming": re.compile(
                r"^(?:\d+\.\d+\% \()?(\d+)(?:\))?(?:.*[of]?)reads\swere(?: quality trimmed | shortened)"
            ),
        }
        primer_regex = re.compile(r"^(.*)(?:\t+)(\d+$)")
        parsed_data = dict()
        primers = dict()
        for line in f["f"]:
            # Search count regexes for stats
            for k, count_regex in count_regexes.items():
                count_match = count_regex.search(line)
                if count_match:
                    parsed_data[k] = int(count_match.group(1))

            # Try to match the primer regex
            primer_match = primer_regex.search(line)
            if primer_match:
                primers[primer_match.group(1)] = int(primer_match.group(2))

        if parsed_data is not None and len(parsed_data) > 0:
            if f["s_name"] in self.ivar_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.ivar_data[f["s_name"]] = parsed_data
            self.add_data_source(f, section="trimming")

        if primers is not None and len(primers) > 0:
            self.ivar_primers[f["s_name"]] = primers
            self.add_data_source(f, section="trimming")

    # Add to general stats table
    def ivar_general_stats_table(self):
        """Take the parsed stats from the iVAR report and add it to the
        basic stats table"""

        headers = {
            "reads_too_short_after_trimming": {
                "title": f"{config.read_count_prefix} Too short",
                "description": "Number of reads too short (<30bp) after primer trimming ({})".format(
                    config.read_count_desc
                ),
                "scale": "OrRd",
                "shared_key": "read_counts",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "reads_outside_primer_region": {
                "title": f"{config.read_count_prefix} Outside primer",
                "description": f"Number of reads outside the primer region ({config.read_count_desc})",
                "scale": "YlOrBr",
                "shared_key": "read_counts",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "trimmed_reads": {
                "title": f"{config.read_count_prefix} Primer trimmed",
                "description": "Total number of reads where primer trimming was performed. ({})".format(
                    config.read_count_desc
                ),
                "scale": "Purples",
                "shared_key": "read_counts",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "mapped_reads": {
                "title": f"{config.read_count_prefix} Mapped",
                "description": f"Total number of mapped reads in iVar input. ({config.read_count_desc})",
                "scale": "PuBu",
                "shared_key": "read_counts",
                "modify": lambda x: x * config.read_count_multiplier,
            },
        }
        self.general_stats_addcols(self.ivar_data, headers)

    def primer_heatmap(self):
        """Heatmap showing information on each primer found for every sample"""
        # Top level dict contains sample IDs + dict(primer, counts)

        final_data = []
        final_xcats = []
        final_ycats = []

        for i, (k, v) in enumerate(self.ivar_primers.items()):
            final_ycats.append(k)
            tmp_prim_val = []
            for prim, val in v.items():
                if i == 0:
                    final_xcats.append(prim)
                tmp_prim_val.append(val)
            final_data.append(tmp_prim_val)

        if len(final_xcats) != len(final_data[0]):
            log.error(f"Number of primers do not match: {len(final_xcats)} vs {len(final_data[0])}")

        if len(final_ycats) == len(final_data):
            log.error(f"Number of samples and primers do not match: {len(final_ycats)} vs {len(final_data)}")

        if self.ivar_primers is not None:
            pconfig = {
                "id": "ivar-primer-count-heatmap",
                "title": "iVar: Number of primers found for each sample",
                "decimalPlaces": 0,
                "square": False,
                "xcats_samples": False,
            }

            self.add_section(
                name="iVar Primer Counts",
                anchor="ivar-primers-heatmap",
                description="Counts observed for each primer per sample.",
                helptext="This lists the number of times a specific primer was found in the respective sample.",
                plot=heatmap.plot(final_data, final_xcats, final_ycats, pconfig),
            )



""" MultiQC module to parse output from vcftools """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

from .relatedness2 import Relatedness2Mixin
from .tstv_by_count import TsTvByCountMixin
from .tstv_by_qual import TsTvByQualMixin
from .tstv_summary import TsTvSummaryMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, Relatedness2Mixin, TsTvByCountMixin, TsTvByQualMixin, TsTvSummaryMixin):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="VCFTools",
            anchor="vcftools",
            href="https://vcftools.github.io",
            info="is a program for working with and reporting on VCF files.",
            doi="10.1093/bioinformatics/btr330",
        )

        n = dict()
        n["relatedness2"] = self.parse_relatedness2()

        n["tstv_by_count"] = self.parse_tstv_by_count()
        if n["tstv_by_count"] > 0:
            log.info(f"Found {n['tstv_by_count']} TsTv.count reports")

        n["tstv_by_qual"] = self.parse_tstv_by_qual()
        if n["tstv_by_qual"] > 0:
            log.info(f"Found {n['tstv_by_qual']} TsTv.qual reports")

        n["tstv_summary"] = self.parse_tstv_summary()
        if n["tstv_summary"] > 0:
            log.info(f"Found {n['tstv_summary']} TsTv.summary reports")

        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound



#!/usr/bin/env python

""" MultiQC functions to plot a scatter plot """

import logging
import random

from multiqc.utils import config, report

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(plotdata, pconfig=None):
    return highcharts_scatter_plot(plotdata, pconfig)


def highcharts_scatter_plot(plotdata, pconfig=None):
    """
    Build the HTML needed for a HighCharts scatter plot. Should be
    called by scatter.plot(), which properly formats input data.
    """
    if pconfig is None:
        pconfig = {}

    # Get the plot ID
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_hcplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    # Build the HTML for the page
    html = '<div class="mqc_hcplot_plotgroup">'

    # Buttons to cycle through different datasets
    if len(plotdata) > 1:
        html += '<div class="btn-group hc_switch_group">\n'
        for k, p in enumerate(plotdata):
            active = "active" if k == 0 else ""
            dls = [dl if isinstance(dl, dict) else {"name": dl} for dl in pconfig.get("data_labels", [])]
            try:
                name = dls[k]["name"]
            except Exception:
                name = k + 1
            try:
                ylab = f"data-ylab=\"{dls[k]['ylab']}\""
            except Exception:
                ylab = f'data-ylab="{name}"' if name != k + 1 else ""
            try:
                ymax = f"data-ymax=\"{dls[k]['ymax']}\""
            except Exception:
                ymax = ""
            try:
                xlab = f"data-xlab=\"{dls[k]['xlab']}\""
            except Exception:
                xlab = f'data-xlab="{name}"' if name != k + 1 else ""
            html += '<button class="btn btn-default btn-sm {a}" data-action="set_data" {y} {ym} {xl} data-newdata="{k}" data-target="{id}">{n}</button>\n'.format(
                a=active, id=pconfig["id"], n=name, y=ylab, ym=ymax, xl=xlab, k=k
            )
        html += "</div>\n\n"

    # The plot div
    html += '<div class="hc-plot-wrapper"{height}><div id="{id}" class="hc-plot not_rendered hc-scatter-plot"><small>loading..</small></div></div></div> \n'.format(
        id=pconfig["id"],
        height=f' style="height:{pconfig["height"]}px"' if "height" in pconfig else "",
    )

    report.num_hc_plots += 1

    # Reverse order of dots in plotdata as the z-order is reversed in Highcharts compared to Plotly
    for d in plotdata:
        d.reverse()

    report.plot_data[pconfig["id"]] = {"plot_type": "scatter", "datasets": plotdata, "config": pconfig}

    return html



""" MultiQC module to parse output from pychopper """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Pychopper",
            anchor="pychopper",
            href="https://github.com/nanoporetech/pychopper",
            info="is a tool to identify, orient, trim and rescue full length Nanopore cDNA reads.",
            # Can't find a DOI // doi=
        )

        # Parse stats file
        self.pychopper_data = {}
        for f in self.find_log_files("pychopper"):
            sample = f["s_name"]
            self.pychopper_data[sample] = {}
            lines = f["f"].splitlines()
            for line in lines[1:]:
                category, name, value = line.split()
                if category not in self.pychopper_data[sample]:
                    self.pychopper_data[sample][category] = {}
                self.pychopper_data[sample][category][name] = float(value)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.pychopper_data = self.ignore_samples(self.pychopper_data)

        # Raise user warning if no data found
        if len(self.pychopper_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.pychopper_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Add to general statistics table:
        # Percentage of full length transcripts
        data_general_stats = {}
        for sample in self.pychopper_data.keys():
            data_general_stats[sample] = {}
            c = self.pychopper_data[sample]["Classification"]
            ftp = c["Primers_found"] * 100 / (c["Primers_found"] + c["Rescue"] + c["Unusable"])
            data_general_stats[sample]["ftp"] = ftp

        headers = {
            "ftp": {
                "title": "Full-Length cDNA",
                "description": "Percentage of full length cDNA reads with correct primers at both ends",
                "suffix": "%",
                "max": 100,
                "min": 0,
            }
        }

        self.general_stats_addcols(data_general_stats, headers)

        # Write data file
        self.write_data_file(self.pychopper_data, "multiqc_pychopper")

        # Report sections
        self.add_section(
            name="cDNA Read Classification",
            description=(
                """
                This plot shows the cDNA read categories identified by Pychopper </br>
                """
            ),
            helptext=(
                """
                There are three possible cases:

                * **Primers found**: Full length cDNA reads with correct primers at both ends.
                * **Rescued reads**: Split fusion reads.
                * **Unusable**: Reads without correct primer combinations.
                """
            ),
            anchor="pychopper_classification",
            plot=self.plot_classification(),
        )
        self.add_section(
            name="cDNA Strand Orientation",
            description=(
                """
                This plot shows the strand orientation of full length cDNA reads
                """
            ),
            helptext=(
                """
                Nanopore cDNA reads are always read forward. To estimate their original strand,
                Pychopper searches for the location of the start and end primers and assigns the reads accordingly.
                """
            ),
            anchor="pychopper_orientation",
            plot=self.plot_orientation(),
        )

    # Plotting functions
    def plot_classification(self):
        """Generate the cDNA read classification plot"""

        pconfig = {
            "id": "pychopper_classification_plot",
            "title": "Pychopper: Read classification",
            "ylab": "",
            "xDecimals": False,
            "ymin": 0,
        }

        data_classification = {}
        for sample in self.pychopper_data.keys():
            data_classification[sample] = {}
            data_classification[sample] = self.pychopper_data[sample]["Classification"]

        cats = ["Primers_found", "Rescue", "Unusable"]
        return bargraph.plot(data_classification, cats, pconfig)

    def plot_orientation(self):
        """Generate the read strand orientation plot"""

        pconfig = {
            "id": "pychopper_orientation_plot",
            "title": "Pychopper: Strand Orientation",
            "ylab": "",
            "cpswitch_c_active": False,
            "xDecimals": False,
            "ymin": 0,
        }

        data_orientation = {}
        for sample in self.pychopper_data.keys():
            data_orientation[sample] = {}
            data_orientation[sample] = self.pychopper_data[sample]["Strand"]

        cats = ["+", "-"]
        return bargraph.plot(data_orientation, cats, pconfig)



from itertools import chain

from multiqc.plots import linegraph


def plot_qchist(samples, file_type, **plot_args):
    """Create line graph plot of histogram data for BBMap 'qchist' output.

    The 'samples' parameter could be from the bbmap mod_data dictionary:
    samples = bbmap.MultiqcModule.mod_data[file_type]
    """

    sumy = sum([int(samples[sample]["data"][x][0]) for sample in samples for x in samples[sample]["data"]])

    cutoff = sumy * 0.999
    all_x = set()
    for item in sorted(chain(*[samples[sample]["data"].items() for sample in samples])):
        all_x.add(item[0])
        cutoff -= item[1][0]
        if cutoff < 0:
            xmax = item[0]
            break
    else:
        xmax = max(all_x)

    data = {
        sample: {x: samples[sample]["data"][x][0] if x in samples[sample]["data"] else 0 for x in all_x}
        for sample in samples
    }
    # Add a count of 0.1 to zero counts, to avoid broken series in log axis
    data = {s: {k: d + 0.1 if d == 0 else d for k, d in v.items()} for s, v in data.items()}

    plot_params = {
        "id": "bbmap-" + file_type + "_plot",
        "title": "BBTools: " + plot_args["plot_title"],
        "xmax": xmax,
    }
    plot_params.update(plot_args["plot_params"])
    plot = linegraph.plot(data, plot_params)

    return plot



""" MultiQC module to parse output from Kallisto """


import logging
import os
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Kallisto module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Kallisto",
            anchor="kallisto",
            href="http://pachterlab.github.io/kallisto/",
            info="is a program for quantifying abundances of transcripts from RNA-Seq data.",
            doi="10.1038/nbt.3519",
        )

        # Find and load any Kallisto reports
        self.kallisto_data = dict()
        for f in self.find_log_files("kallisto", filehandles=True):
            self.parse_kallisto_log(f)

        # Filter to strip out ignored sample names
        self.kallisto_data = self.ignore_samples(self.kallisto_data)

        if len(self.kallisto_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.kallisto_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.kallisto_data, "multiqc_kallisto")

        # Basic Stats Table
        self.kallisto_general_stats_table()

        # Alignment Rate Plot
        self.add_section(plot=self.kallisto_alignment_plot())

    def parse_kallisto_log(self, f):
        s_name = total_reads = paligned_reads = fraglength = None
        for line in f["f"]:
            # Get input filename
            match = re.search(r"\[quant\] will process (pair|file|sample) 1: (\S+)", line)
            if match:
                s_name = self.clean_s_name(os.path.basename(match.group(2)), f)

            if s_name is not None:
                # Alignment rates
                aligned = re.search(r"\[quant\] processed ([\d,]+) reads, ([\d,]+) reads pseudoaligned", line)
                if aligned:
                    total_reads = float(aligned.group(1).replace(",", ""))
                    paligned_reads = float(aligned.group(2).replace(",", ""))

                # Paired end fragment lengths
                flength = re.search(r"\[quant\] estimated average fragment length: ([\d\.]+)", line)
                if flength:
                    fraglength = float(flength.group(1).replace(",", ""))

                if "quantifying the abundances" in line:
                    if s_name in self.kallisto_data:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.kallisto_data[s_name] = {
                        "total_reads": total_reads,
                        "pseudoaligned_reads": paligned_reads,
                        "not_pseudoaligned_reads": total_reads - paligned_reads,
                    }
                    try:
                        self.kallisto_data[s_name]["percent_aligned"] = (paligned_reads / total_reads) * 100
                    except ZeroDivisionError:
                        self.kallisto_data[s_name]["percent_aligned"] = 0.0
                    if fraglength is not None:
                        self.kallisto_data[s_name]["fragment_length"] = fraglength
                    s_name = total_reads = paligned_reads = fraglength = None

    def kallisto_general_stats_table(self):
        """Take the parsed stats from the Kallisto report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "fragment_length": {
                "title": "Frag Length",
                "description": "Estimated average fragment length",
                "min": 0,
                "suffix": "bp",
                "scale": "RdYlGn",
            },
            "percent_aligned": {
                "title": "% Aligned",
                "description": "% processed reads that were pseudoaligned",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "pseudoaligned_reads": {
                "title": f"{config.read_count_prefix} Aligned",
                "description": f"Pseudoaligned reads ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.kallisto_data, headers)

    def kallisto_alignment_plot(self):
        """Make the HighCharts HTML to plot the alignment rates"""

        # Specify the order of the different possible categories
        keys = {
            "pseudoaligned_reads": {"color": "#437bb1", "name": "Pseudoaligned"},
            "not_pseudoaligned_reads": {"color": "#b1084c", "name": "Not aligned"},
        }

        # Config for the plot
        config = {
            "id": "kallisto_alignment",
            "title": "Kallisto: Alignment Scores",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.kallisto_data, keys, config)



# Initialise the logger
import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph

log = logging.getLogger(__name__)


class DragenRnaTranscriptCoverage(BaseMultiqcModule):
    def add_rna_transcript_coverage(self):
        data_by_sample = defaultdict(dict)

        for f in self.find_log_files("dragen/rna_transcript_cov"):
            data = parse_rna_transcript_cov(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="rna_transcript_cov")
            data_by_sample[s_name] = data

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)

        if not data_by_sample:
            return set()

        # Only plot data, don't want to write this to a file
        # (can do so with --export-plots already)
        # self.write_data_file(data_by_sample, "dragen_transcript_cov")

        self.add_section(
            name="RNA Transcript Coverage",
            anchor="rna-transcript-cov",
            description="""
            RNA transcript coverage.  This is the average coverage at the position along the transcripts.
            """,
            plot=linegraph.plot(
                data_by_sample,
                pconfig={
                    "id": "dragen_rna_transcript_cov",
                    "title": "Dragen: RNA Transcript Coverage",
                    "ylab": "Average coverage",
                    "xlab": "Transcript position",
                    "categories": True,
                    "tt_label": "<b>{point.x}</b>: {point.y:.1f}x",
                },
            ),
        )

        return data_by_sample.keys()


def parse_rna_transcript_cov(f):
    """
    *.quant.transcript_coverage.txt
    Percentile	Coverage
    1	0.0518503
    2	0.125753
    3	0.194324
    4	0.254819
    5	0.308795
    """

    data = {}
    for line in f["f"].splitlines()[1:]:
        percentile, coverage = line.split()

        try:
            percentile = float(percentile)
        except ValueError:
            pass
        try:
            coverage = float(coverage)
        except ValueError:
            pass
        data[percentile] = coverage

    return data



import copy
import dataclasses
import logging
from collections import defaultdict
from typing import Dict, List, Union, Optional

import numpy as np
from plotly import graph_objects as go

from multiqc.plots.plotly.plot import Plot, PlotType, BaseDataset
from multiqc.utils import util_functions

logger = logging.getLogger(__name__)


# {'color': 'rgb(211,211,211,0.05)', 'name': 'background: EUR', 'x': -0.294, 'y': -1.527}
PointT = Dict[str, Union[str, float, int]]


def plot(points_lists: List[List[PointT]], pconfig: Dict) -> str:
    """
    Build and add the plot data to the report, return an HTML wrapper.
    :param points_lists: each dataset is a 2D dict, first keys as sample names, then x:y data pairs
    :param pconfig: dict with config key:value pairs. See CONTRIBUTING.md
    :return: HTML with JS, ready to be inserted into the page
    """
    p = ScatterPlot(pconfig, points_lists)

    from multiqc.utils import report

    return p.add_to_report(report)


class ScatterPlot(Plot):
    @dataclasses.dataclass
    class Dataset(BaseDataset):
        points: List[PointT]
        categories: List[str]

        @staticmethod
        def create(
            dataset: BaseDataset,
            points: List[Dict],
            pconfig: Dict,
        ) -> "ScatterPlot.Dataset":
            dataset = ScatterPlot.Dataset(
                **dataset.__dict__,
                points=points,
                categories=pconfig.get("categories", []),
            )

            dataset.trace_params.update(
                textfont=dict(size=8),
                marker=dict(
                    size=10,
                    line=dict(width=1),
                    opacity=1,
                    color="rgba(124, 181, 236, .5)",
                ),
            )
            return dataset

        def create_figure(
            self,
            layout: Optional[go.Layout] = None,
            is_log=False,
            is_pct=False,
        ) -> go.Figure:
            """
            Create a Plotly figure for a dataset
            """
            fig = go.Figure(layout=layout)

            MAX_ANNOTATIONS = 10  # Maximum number of dots to be annotated directly on the plot
            n_annotated = len([el for el in self.points if "annotation" in el])
            if n_annotated < MAX_ANNOTATIONS:
                # Finding and marking outliers to only label them
                # 1. Calculate Z-scores
                points = [(i, x) for i, x in enumerate(self.points) if x.get("annotate", True) is not False]
                x_values = np.array([x["x"] for (i, x) in points])
                y_values = np.array([x["y"] for (i, x) in points])
                x_std = np.std(x_values)
                y_std = np.std(y_values)
                if x_std == 0 and y_std == 0:
                    logger.warning(f"Scatter plot {self.plot.id}: all {len(points)} points have the same coordinates")
                    if len(points) == 1:  # Only single point - annotate it!
                        for i, point in points:
                            point["annotation"] = point["name"]
                else:
                    x_z_scores = np.abs((x_values - np.mean(x_values)) / x_std) if x_std else np.zeros_like(x_values)
                    y_z_scores = np.abs((y_values - np.mean(y_values)) / y_std) if y_std else np.zeros_like(y_values)
                    # 2. Find a reasonable threshold so there are not too many outliers
                    threshold = 1.0
                    while threshold <= 6.0:
                        n_outliers = np.count_nonzero((x_z_scores > threshold) | (y_z_scores > threshold))
                        logger.debug(f"Scatter plot outlier threshold: {threshold:.2f}, outliers: {n_outliers}")
                        if n_annotated + n_outliers <= MAX_ANNOTATIONS:
                            break
                        # If there are too many outliers, we increase the threshold until we have less than 10
                        threshold += 0.2
                    # 3. Annotate outliers that pass the threshold
                    for (i, point), x_z_score, y_z_score in zip(points, x_z_scores, y_z_scores):
                        # Check if point is an outlier or if total points are less than 10
                        if x_z_score > threshold or y_z_score > threshold:
                            point["annotation"] = point["name"]
                n_annotated = len([point for point in self.points if "annotation" in point])

            # If there are few unique colors, we can additionally put a unique list into a legend
            # (even though some color might belong to many distinct names - we will just crop the list)
            names_by_legend_key = defaultdict(set)
            for el in self.points:
                legend_key = (el.get("color"), el.get("marker_size"), el.get("marker_line_width"), el.get("group"))
                names_by_legend_key[legend_key].add(el["name"])
            layout.showlegend = True

            in_legend = set()
            for element in self.points:
                x = element["x"]
                if self.categories:
                    if isinstance(x, float):
                        x = int(round(x))
                    if isinstance(x, int) and (0 <= x < len(self.categories)):
                        x = self.categories[x]
                    else:
                        logger.error(
                            f"Scatter plot {self.plot.id}: x={x} must be an index in the list of categories: {self.categories}"
                        )
                        continue

                name = element["name"]
                group = element.get("group")
                color = element.get("color")
                annotation = element.get("annotation")

                show_in_legend = False
                if layout.showlegend and not element.get("hide_in_legend"):
                    key = (color, element.get("marker_size"), element.get("marker_line_width"), group)
                    if key not in in_legend:
                        in_legend.add(key)
                        names = sorted(names_by_legend_key.get(key))
                        label = ", ".join(names)
                        if group:
                            label = f"{group}: {label}"
                        MAX_WIDTH = 60
                        if len(label) > MAX_WIDTH:  # Crop long labels
                            label = label[:MAX_WIDTH] + "..."
                        show_in_legend = True
                        name = label

                params = copy.deepcopy(self.trace_params)
                marker = params.pop("marker")
                if color:
                    marker["color"] = color
                if "marker_line_width" in element:
                    marker["line"]["width"] = element["marker_line_width"]
                if "marker_size" in element:
                    marker["size"] = element["marker_size"]
                if "opacity" in element:
                    marker["opacity"] = element["opacity"]

                if annotation:
                    params["mode"] = "markers+text"
                if n_annotated > 0:  # Reduce opacity of the borders that clutter the annotations:
                    marker["line"]["color"] = "rgba(0, 0, 0, .2)"

                fig.add_trace(
                    go.Scatter(
                        x=[x],
                        y=[element["y"]],
                        name=name,
                        text=[annotation or name],
                        showlegend=show_in_legend,
                        marker=marker,
                        **params,
                    )
                )
            fig.layout.height += len(in_legend) * 5  # extra space for legend
            return fig

    def __init__(self, pconfig: Dict, points_lists: List[List[PointT]]):
        super().__init__(PlotType.SCATTER, pconfig, len(points_lists))

        self.datasets: List[ScatterPlot.Dataset] = [
            ScatterPlot.Dataset.create(d, points, pconfig) for d, points in zip(self.datasets, points_lists)
        ]

        # Make a tooltip always show on hover over nearest point on plot
        self.layout.hoverdistance = -1

    def tt_label(self) -> Optional[str]:
        """Default tooltip label"""
        return "<br><b>X</b>: %{x}<br><b>Y</b>: %{y}"

    def save_data_file(self, dataset: Dataset) -> None:
        data = [
            {
                "Name": point["name"],
                "X": point["x"],
                "Y": point["y"],
            }
            for point in dataset.points
        ]
        util_functions.write_data_file(data, dataset.uid)



""" MultiQC module to parse output from Afterqc """


import json
import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    AfterQC module class
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="AfterQC",
            anchor="afterqc",
            href="https://github.com/OpenGene/AfterQC",
            info="Automatic Filtering, Trimming, Error Removing and Quality Control for fastq data.",
            doi="10.1186/s12859-017-1469-3",
        )

        # Find and load any Afterqc reports
        self.afterqc_data = dict()
        for f in self.find_log_files("afterqc", filehandles=True):
            self.parse_afterqc_log(f)

        # Filter to strip out ignored sample names
        self.afterqc_data = self.ignore_samples(self.afterqc_data)

        if len(self.afterqc_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.afterqc_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.afterqc_data, "multiqc_afterqc")

        # Basic Stats Table
        self.afterqc_general_stats_table()

        # Alignment bar plot
        self.add_section(
            name="Bad Reads",
            anchor="after_qc",
            description="Filtering statistics of sampled reads.",
            plot=self.after_qc_bad_reads_chart(),
        )

    def parse_afterqc_log(self, f):
        """Parse the JSON output from AfterQC and save the summary statistics"""
        try:
            parsed_json = json.load(f["f"])
        except Exception:
            log.warning(f"Could not parse AfterQC JSON: '{f['fn']}'")
            return None

        # AfterQC changed the name of their summary key at some point
        if "summary" in parsed_json:
            summaryk = "summary"
        elif "afterqc_main_summary" in parsed_json:
            summaryk = "afterqc_main_summary"
        else:
            log.warning(f"AfterQC JSON did not have a 'summary' or 'afterqc_main_summary' key, skipping: '{f['fn']}'")
            return None

        s_name = f["s_name"]
        self.add_data_source(f, s_name)
        self.afterqc_data[s_name] = {}
        for k in parsed_json[summaryk]:
            try:
                self.afterqc_data[s_name][k] = float(parsed_json[summaryk][k])
            except ValueError:
                self.afterqc_data[s_name][k] = parsed_json[summaryk][k]
        try:
            self.afterqc_data[s_name]["pct_good_bases"] = (
                self.afterqc_data[s_name]["good_bases"] / self.afterqc_data[s_name]["total_bases"]
            ) * 100.0
        except KeyError:
            pass

    def afterqc_general_stats_table(self):
        """Take the parsed stats from the Afterqc report and add it to the
        General Statistics table at the top of the report"""

        headers = {
            "pct_good_bases": {
                "title": "% Good Bases",
                "description": "Percent Good Bases",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "BuGn",
            },
            "good_reads": {
                "title": f"{config.read_count_prefix} Good Reads",
                "description": f"Good Reads ({config.read_count_desc})",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
                "scale": "GnBu",
                "shared_key": "read_count",
            },
            "total_reads": {
                "title": f"{config.read_count_prefix} Total Reads",
                "description": f"Total Reads ({config.read_count_desc})",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
                "scale": "Blues",
                "shared_key": "read_count",
            },
            "readlen": {
                "title": "Read Length",
                "description": "Read Length",
                "min": 0,
                "suffix": " bp",
                "format": "{:,.0f}",
                "scale": "YlGn",
            },
        }
        self.general_stats_addcols(self.afterqc_data, headers)

    def after_qc_bad_reads_chart(self):
        """Function to generate the AfterQC bad reads bar plot"""
        # Specify the order of the different possible categories
        keys = {
            "good_reads": {"name": "Good Reads"},
            "bad_reads_with_bad_barcode": {"name": "Bad Barcode"},
            "bad_reads_with_bad_overlap": {"name": "Bad Overlap"},
            "bad_reads_with_bad_read_length": {"name": "Bad Read Length"},
            "bad_reads_with_low_quality": {"name": "Low Quality"},
            "bad_reads_with_polyX": {"name": "PolyX"},
            "bad_reads_with_reads_in_bubble": {"name": "Reads In Bubble"},
            "bad_reads_with_too_many_N": {"name": "Too many N"},
        }

        # Config for the plot
        pconfig = {
            "id": "afterqc_bad_reads_plot",
            "title": "AfterQC: Filtered Reads",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
            "hide_zero_cats": False,
        }
        return bargraph.plot(self.afterqc_data, keys, pconfig)



""" MultiQC module to parse output from QUAST """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="QUAST",
            anchor="quast",
            href="http://quast.bioinf.spbau.ru/",
            info="is a quality assessment tool for genome assemblies, written by the Center for Algorithmic Biotechnology.",
            doi="10.1093/bioinformatics/btt086",
        )

        # Get modifiers from config file
        qconfig = getattr(config, "quast_config", {})

        self.contig_length_multiplier = qconfig.get("contig_length_multiplier", 0.001)
        self.contig_length_suffix = qconfig.get("contig_length_suffix", "Kbp")

        self.total_length_multiplier = qconfig.get("total_length_multiplier", 0.000001)
        self.total_length_suffix = qconfig.get("total_length_suffix", "Mbp")

        self.total_number_contigs_multiplier = qconfig.get("total_number_contigs_multiplier", 0.001)
        self.total_number_contigs_suffix = qconfig.get("total_number_contigs_suffix", "K")

        # Find and load any QUAST reports
        self.quast_data = {}
        for f in self.find_log_files("quast"):
            self.parse_quast_log(f)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        # Filter to strip out ignored sample names
        self.quast_data = self.ignore_samples(self.quast_data)

        if len(self.quast_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.quast_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.quast_data, "multiqc_quast")

        # Basic Stats Table
        self.quast_general_stats_table()

        # Quast Stats Table
        self.add_section(name="Assembly Statistics", anchor="quast-stats", plot=self.quast_table())
        # Number of contigs plot
        self.add_section(
            name="Number of Contigs",
            anchor="quast-contigs",
            description="""This plot shows the number of contigs found for each assembly, broken
                    down by length.""",
            plot=self.quast_contigs_barplot(),
        )
        # Number of genes plot
        ng_pdata = self.quast_predicted_genes_barplot()
        if ng_pdata:
            self.add_section(
                name="Number of Predicted Genes",
                anchor="quast-genes",
                description="""This plot shows the number of predicted genes found for each
                          assembly, broken down by length.""",
                plot=ng_pdata,
            )
        # Number of partial genes plot
        ng_pdata = self.quast_predicted_genes_barplot(partial=True)
        if ng_pdata:
            self.add_section(
                name="Number of Partially Predicted Genes",
                anchor="quast-partial-genes",
                description="""This plot shows the number of partially predicted genes found for each
                          assembly, broken down by length.""",
                plot=ng_pdata,
            )

    def parse_quast_log(self, f):
        lines = f["f"].splitlines()

        # Pull out the sample names from the first row
        s_names = lines[0].split("\t")
        # Prepend directory name(s) to sample names as configured
        s_names = [self.clean_s_name(s_name, f) for s_name in s_names]
        for s_name in s_names[1:]:
            if s_name in self.quast_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.quast_data[s_name] = dict()

        # Parse remaining stats for each sample
        for line in lines[1:]:
            s = line.split("\t")
            k = s[0]
            for i, v in enumerate(s[1:]):
                s_name = s_names[i + 1]
                partials = re.search(r"(\d+) \+ (\d+) part", v)
                if partials:
                    whole = partials.group(1)
                    partial = partials.group(2)
                    try:
                        self.quast_data[s_name][k] = float(whole)
                        self.quast_data[s_name][f"{k}_partial"] = float(partial)
                    except ValueError:
                        self.quast_data[s_name][k] = whole
                        self.quast_data[s_name][f"{k}_partial"] = partial
                else:
                    try:
                        self.quast_data[s_name][k] = float(v)
                    except ValueError:
                        self.quast_data[s_name][k] = v

    def quast_general_stats_table(self):
        """Take the parsed stats from the QUAST report and add some to the
        General Statistics table at the top of the report"""

        headers = {
            "N50": {
                "title": f"N50 ({self.contig_length_suffix})",
                "description": "N50 is the contig length such that using longer or equal length contigs produces half (50%) of the bases of the assembly (kilo base pairs)",
                "min": 0,
                "suffix": self.contig_length_suffix,
                "scale": "RdYlGn",
                "modify": lambda x: x * self.contig_length_multiplier,
            },
            "Total length": {
                "title": f"Assembly Length ({self.total_length_suffix})",
                "description": f"The total number of bases in the assembly ({self.total_length_suffix}).",
                "min": 0,
                "suffix": self.total_length_suffix,
                "scale": "YlGn",
                "modify": lambda x: x * self.total_length_multiplier,
            },
        }
        self.general_stats_addcols(self.quast_data, headers)

    def quast_table(self):
        """Write some more statistics about the assemblies in a table."""
        headers = {
            "N50": {
                "title": f"N50 ({self.contig_length_suffix})",
                "description": "N50 is the contig length such that using longer or equal length contigs produces half (50%) of the bases of the assembly.",
                "min": 0,
                "suffix": self.contig_length_suffix,
                "scale": "RdYlGn",
                "modify": lambda x: x * self.contig_length_multiplier,
            },
            "N75": {
                "title": f"N75 ({self.contig_length_suffix})",
                "description": "N75 is the contig length such that using longer or equal length contigs produces 75% of the bases of the assembly",
                "min": 0,
                "suffix": self.contig_length_suffix,
                "scale": "RdYlGn",
                "modify": lambda x: x * self.contig_length_multiplier,
            },
            "L50": {
                "title": f"L50 ({self.total_number_contigs_suffix})" if self.total_number_contigs_suffix else "L50",
                "description": "L50 is the number of contigs larger than N50, i.e. the minimum number of contigs comprising 50% of the total assembly length.",
                "min": 0,
                "suffix": self.total_number_contigs_suffix,
                "scale": "RdYlGn-rev",
                "modify": lambda x: x * self.total_number_contigs_multiplier,
            },
            "L75": {
                "title": f"L75 ({self.total_number_contigs_suffix})" if self.total_number_contigs_suffix else "L75",
                "description": "L75 is the number of contigs larger than N75, i.e. the minimum number of contigs comprising 75% of the total assembly length.",
                "min": 0,
                "suffix": self.total_number_contigs_suffix,
                "scale": "RdYlGn-rev",
                "modify": lambda x: x * self.total_number_contigs_multiplier,
            },
            "Largest contig": {
                "title": f"Largest contig ({self.contig_length_suffix})",
                "description": "The size of the largest contig of the assembly",
                "min": 0,
                "suffix": self.contig_length_suffix,
                "scale": "YlGn",
                "modify": lambda x: x * self.contig_length_multiplier,
            },
            "Total length": {
                "title": f"Length ({self.total_length_suffix})",
                "description": "The total number of bases in the assembly.",
                "min": 0,
                "suffix": self.total_length_suffix,
                "scale": "YlGn",
                "modify": lambda x: x * self.total_length_multiplier,
            },
            "# misassemblies": {
                "title": "Misassemblies",
                "description": "The number of positions in the assembled contigs where the left flanking sequence aligns over 1 kbp away from the right flanking sequence on the reference (relocation) or they overlap on more than 1 kbp (relocation) or flanking sequences align on different strands (inversion) or different chromosomes (translocation).",
                "scale": "RdYlGn-rev",
                "format": "{,:.0f}",
            },
            "# mismatches per 100 kbp": {
                "title": "Mismatches/100kbp",
                "description": "The number of mismatches per 100 kbp",
                "scale": "YlOrRd",
                "format": "{:,.2f}",
            },
            "# indels per 100 kbp": {
                "title": "Indels/100kbp",
                "description": "The number of indels per 100 kbp",
                "scale": "YlOrRd",
                "format": "{:,.2f}",
            },
            "# genes": {
                "title": "Genes",
                "description": "# Genes",
                "scale": "YlGnBu",
                "format": "{:,.0f}",
                "shared_key": "gene_count",
            },
            "# genes_partial": {
                "title": "Genes (Partial)",
                "description": "# Genes (Partial)",
                "scale": "YlGnBu",
                "format": "{:,.0f}",
                "shared_key": "gene_count",
            },
            "# predicted genes (unique)": {
                "title": "Genes",
                "description": "# Predicted Genes (Unique)",
                "scale": "YlGnBu",
                "format": "{:,.0f}",
                "shared_key": "gene_count",
            },
            "Genome fraction (%)": {
                "title": "Genome Fraction",
                "description": "The total number of aligned bases in the reference, divided by the genome size.",
                "max": 100,
                "suffix": "%",
                "scale": "YlGn",
            },
        }

        config = {
            "id": "quast_table",
            "min": 0,
            "namespace": "QUAST",
            "title": "QUAST: Assembly Statistics",
        }
        return table.plot(self.quast_data, headers, config)

    def quast_contigs_barplot(self):
        """Make a bar plot showing the number and length of contigs for each assembly"""

        # Prep the data
        data = dict()
        categories = []
        for s_name, d in self.quast_data.items():
            nums_by_t = dict()
            for k, v in d.items():
                m = re.match(r"# contigs \(>= (\d+) bp\)", k)
                if m and v != "-":
                    nums_by_t[int(m.groups()[0])] = int(v)

            tresholds = sorted(nums_by_t.keys(), reverse=True)
            p = dict()
            cats = []
            for i, t in enumerate(tresholds):
                if i == 0:
                    c = ">= " + str(t) + " bp"
                    cats.append(c)
                    p[c] = nums_by_t[t]
                else:
                    c = str(t) + "-" + str(tresholds[i - 1]) + " bp"
                    cats.append(c)
                    p[c] = nums_by_t[t] - nums_by_t[tresholds[i - 1]]
            if not categories:
                categories = cats
            data[s_name] = p

        pconfig = {
            "id": "quast_num_contigs",
            "title": "QUAST: Number of Contigs",
            "ylab": "# Contigs",
            "yDecimals": False,
        }

        return bargraph.plot(data, categories, pconfig)

    def quast_predicted_genes_barplot(self, partial=False):
        """
        Make a bar plot showing the number and length of predicted genes
        for each assembly
        """

        # Prep the data
        # extract the ranges given to quast with "--gene-thresholds"
        # keys look like:
        #   `# predicted genes (>= 300 bp)`
        #   `# predicted genes (>= 300 bp)_partial`
        pattern = re.compile(r"# predicted genes \(>= (\d+) bp\)" + ("_partial" if partial else ""))

        data = {}
        all_categories = []
        data_key = "# predicted genes (>= {} bp)" + ("_partial" if partial else "")
        for s_name, d in self.quast_data.items():
            thresholds = []
            for k in d.keys():
                m = re.match(pattern, k)
                if m:
                    thresholds.append(int(m.groups()[0]))
            thresholds = sorted(list(set(thresholds)))
            if len(thresholds) < 2:
                continue

            highest_threshold = thresholds[-1]
            highest_cat = (highest_threshold, f">= {highest_threshold} bp")  # tuple (key-for-sorting, label)
            all_categories.append(highest_cat)
            plot_data = {highest_cat[1]: d[data_key.format(highest_threshold)]}

            # converting >=T1, >=T2,.. into 0-T1, T1-T2,..
            for low, high in zip(thresholds[:-1], thresholds[1:]):
                cat = (low, f"{low}-{high} bp")
                all_categories.append(cat)
                plot_data[cat[1]] = d[data_key.format(low)] - d[data_key.format(high)]

            data[s_name] = plot_data

        all_categories = [label for k, label in sorted(list(set(all_categories)))]

        if len(all_categories) > 0:
            return bargraph.plot(
                data,
                all_categories,
                {
                    "id": "quast_" + ("partially_" if partial else "") + "predicted_genes",
                    "title": "QUAST: Number of " + ("partially " if partial else "") + "predicted genes",
                    "ylab": "Number of " + ("partially " if partial else "") + "predicted genes",
                },
            )
        else:
            return None



""" MultiQC module to parse output from gffcompare """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, scatter

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"# gffcompare v([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="GffCompare",
            anchor="gffcompare",
            href="https://ccb.jhu.edu/software/stringtie/gffcompare.shtml",
            info="is a tool to compare, merge and annotate one or more GFF files with a reference annotation in GFF format.",
            doi="10.12688/f1000research.23297.1",
        )

        # Parse stats file
        # Everything is hardcoded with linenumbers, needs to be adjusted if output should change.
        # Other versions of GffCompare with different ouptut formats might not work.

        # TODO (@ewels): It would be nice to rewrite this with regular expressions to make it more robust
        # However, for now I'm ok with it as the output is written by the tool and not stdout / stderr,
        # so the line numbering *should* be fairly portable across systems, at least for the same version of the tool.
        # If you come here because the tool has updated and the module has broken, please refactor the below code.

        self.gffcompare_data = {}
        for f in self.find_log_files("gffcompare"):
            self.add_data_source(f)
            sample = f["s_name"]
            self.gffcompare_data[sample] = {}
            lines = f["f"].splitlines()

            # Version info
            version_match = re.search(VERSION_REGEX, lines[0])
            if version_match:
                self.add_software_version(version_match.group(1), sample)

            ## Transcript and loci numbers:
            # Query
            c = [int(s) for s in lines[5].replace("(", " ").split() if s.isdigit()]
            self.gffcompare_data[sample]["counts"] = {}
            self.gffcompare_data[sample]["counts"]["query"] = {}
            self.gffcompare_data[sample]["counts"]["query"]["mRNA"] = c[0]
            self.gffcompare_data[sample]["counts"]["query"]["loci"] = c[1]
            self.gffcompare_data[sample]["counts"]["query"]["multi_exon_transcripts"] = c[2]
            c = [int(s) for s in lines[6].replace("(", " ").split() if s.isdigit()]
            self.gffcompare_data[sample]["counts"]["query"]["multi_transcript_loci"] = c[0]

            # Reference
            c = [int(s) for s in lines[7].replace("(", " ").split() if s.isdigit()]
            self.gffcompare_data[sample]["counts"]["reference"] = {}
            self.gffcompare_data[sample]["counts"]["reference"]["mRNA"] = c[0]
            self.gffcompare_data[sample]["counts"]["reference"]["loci"] = c[1]
            self.gffcompare_data[sample]["counts"]["reference"]["multi_exon_transcripts"] = c[2]
            c = [int(s) for s in lines[8].replace("(", " ").split() if s.isdigit()]
            self.gffcompare_data[sample]["counts"]["super_loci"] = c[0]

            ## Accuracy metrics (Sensitivity/Precision)
            self.gffcompare_data[sample]["accuracy"] = {}
            self.gffcompare_data[sample]["missed"] = {}
            self.gffcompare_data[sample]["novel"] = {}
            for line in lines[10:16]:
                split = line.replace("|", "").replace("Intron chain", "Intron_chain").split()
                self.gffcompare_data[sample]["accuracy"][split[0]] = {}
                self.gffcompare_data[sample]["accuracy"][split[0]]["sensitivity"] = float(split[2])
                self.gffcompare_data[sample]["accuracy"][split[0]]["precision"] = float(split[3])

            ## Additional count data
            self.gffcompare_data[sample]["counts"]["matching_intron_chains"] = [
                int(s) for s in lines[17].replace("(", " ").split() if s.isdigit()
            ][0]
            self.gffcompare_data[sample]["counts"]["matching_transcripts"] = [
                int(s) for s in lines[18].replace("(", " ").split() if s.isdigit()
            ][0]
            self.gffcompare_data[sample]["counts"]["matching_loci"] = [
                int(s) for s in lines[19].replace("(", " ").split() if s.isdigit()
            ][0]

            ## Missed/Novel genomic elements
            self.gffcompare_data[sample]["missed"]["Exons"] = [
                int(s) for s in lines[21].replace("/", " ").split() if s.isdigit()
            ]
            self.gffcompare_data[sample]["novel"]["Exons"] = [
                int(s) for s in lines[22].replace("/", " ").split() if s.isdigit()
            ]
            self.gffcompare_data[sample]["missed"]["Introns"] = [
                int(s) for s in lines[23].replace("/", " ").split() if s.isdigit()
            ]
            self.gffcompare_data[sample]["novel"]["Introns"] = [
                int(s) for s in lines[24].replace("/", " ").split() if s.isdigit()
            ]
            self.gffcompare_data[sample]["missed"]["Loci"] = [
                int(s) for s in lines[25].replace("/", " ").split() if s.isdigit()
            ]
            self.gffcompare_data[sample]["novel"]["Loci"] = [
                int(s) for s in lines[26].replace("/", " ").split() if s.isdigit()
            ]

        # Filter to strip out ignored sample names
        self.gffcompare_data = self.ignore_samples(self.gffcompare_data)

        # Raise user warning if no data found
        if len(self.gffcompare_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.gffcompare_data)} reports")

        # Add nothing to general statistics table

        # Write data file
        # TODO (@ewels) - because the data structure is hierarchical, the default tsv output format gives an empty file
        # If / when refactoring the above parsing code, restructure to a flat dict so that this works with tabular output files.
        self.write_data_file(self.gffcompare_data, "multiqc_gffcompare")

        # Report sectionsq
        self.add_section(
            name="Accuracy values",
            description=(
                """
                Displayed are the accuracy values precisiond and sensitivity for different levels of genomic features.
                The metrics are calculated for the comparison of a query and reference GTF file.
                """
            ),
            helptext=(
                """
                Accuracy metrics are calculated as described in [Burset et al. (1996)](http://dx.doi.org/10.1006/geno.1996.0298). Sensitivity is the true positive rate, Precision
                True Positives are query features that agree with features in the reference. The exact definition depends on the feature level:

                - *Base*: True positives are bases reported at the same coordinates.
                - *Exon*: Comparison units are exons that overlap in query and reference with same coordinates.
                - *Intron chain*: True positives are query transcripts for which all introns coordinates match those in the reference.
                - *Transcript*: More stringent then intron chain, all Exon coordinates need to match. Outer exon coordinates (start + end) can vary by 100 bases in default settings
                - *Locus*: Cluster of exons need to match.

                A more in depth description can be found [here](https://ccb.jhu.edu/software/stringtie/gffcompare.shtml#levels).
                """
            ),
            anchor="gffcompare_accuracy",
            plot=self.plot_accuracy(),
        )

        self.add_section(
            name="Novel features",
            description="Number of novel features, present in the query data but not found in the reference annotation.",
            anchor="gffcompare_novel",
            plot=self.plot_novel(),
        )

        self.add_section(
            name="Missing features",
            description="False negative features, which are found in the reference annotation but missed (not present) in the query data.",
            anchor="gffcompare_missed",
            plot=self.plot_missed(),
        )

    # Plotting functions
    def plot_accuracy(self):
        """Generate GffCompare accuracy plot"""

        datasets = ["Base", "Exon", "Intron", "Intron_chain", "Transcript", "Locus"]

        pconfig = {
            "id": "gffcompare_accuracy_plot",
            "title": "GffCompare: Accuracy values",
            "ylab": "Precision",
            "xlab": "Sensitivity",
            "ymin": 0,
            "ymax": 1,
            "xmin": 0,
            "xmax": 1,
            "data_labels": [{"name": x, "ylab": "Precision", "xlab": "Sensitivity"} for x in datasets],
        }

        data_classification = [
            {
                sample: {
                    "x": self.gffcompare_data[sample]["accuracy"][dataset]["sensitivity"] / 100,
                    "y": self.gffcompare_data[sample]["accuracy"][dataset]["precision"] / 100,
                    "name": dataset,
                }
                for sample in self.gffcompare_data.keys()
            }
            for dataset in datasets
        ]

        return scatter.plot(data_classification, pconfig)

    # Plot number of novel and missing transcripts
    def plot_novel(self):
        """Generate GffCompare novel elements plot"""

        cats = [
            {
                "novel": {"name": "Novel features", "color": "#f7a35c"},
                "matching": {"name": "Matching features", "color": "#8bbc21"},
            }
        ] * 3

        datasets = ["Exons", "Introns", "Loci"]
        pconfig = {
            "id": "gffcompare_novel_plot",
            "title": "Gffcompare: Novel features",
            "ylab": "Reads",
            "data_labels": datasets,
        }

        data_novel = [
            {
                sample: {
                    "novel": self.gffcompare_data[sample]["novel"][dataset][0],
                    "matching": (
                        self.gffcompare_data[sample]["novel"][dataset][1]
                        - self.gffcompare_data[sample]["novel"][dataset][0]
                    ),
                }
                for sample in self.gffcompare_data.keys()
            }
            for dataset in datasets
        ]

        return bargraph.plot(data_novel, cats, pconfig)

    def plot_missed(self):
        """Generate GffCompare missed elements plot"""

        cats = [
            {
                "missed": {"name": "Missing features", "color": "#f7a35c"},
                "matching": {"name": "Matching features", "color": "#8bbc21"},
            }
        ] * 3

        datasets = ["Exons", "Introns", "Loci"]
        pconfig = {
            "id": "gffcompare_missing_plot",
            "title": "Gffcompare: Missing features",
            "ylab": "Reads",
            "data_labels": datasets,
        }

        data_missed = [
            {
                sample: {
                    "missed": self.gffcompare_data[sample]["missed"][dataset][0],
                    "matching": (
                        self.gffcompare_data[sample]["missed"][dataset][1]
                        - self.gffcompare_data[sample]["missed"][dataset][0]
                    ),
                }
                for sample in self.gffcompare_data.keys()
            }
            for dataset in datasets
        ]

        return bargraph.plot(data_missed, cats, pconfig)



""" MultiQC module to parse output from optitype """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    optitype module class, parses TSV output from optitype.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="OptiType",
            anchor="optitype",
            href="https://github.com/FRED-2/OptiType",
            info="Precision HLA typing from next-generation sequencing data.",
            doi="10.1093/bioinformatics/btu548",
        )

        # Find and load any optitype reports
        self.optitype_data = dict()

        for f in self.find_log_files("optitype"):
            rows = f["f"].splitlines()
            # First col is empty / or always zero
            headers = rows[0].split("\t")[1:]
            cols = rows[1].split("\t")[1:]

            if f["s_name"] in self.optitype_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.optitype_data[f["s_name"]] = dict()
            for header, col in zip(headers, cols):
                self.optitype_data[f["s_name"]][header] = col
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.optitype_data = self.ignore_samples(self.optitype_data)

        if len(self.optitype_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.optitype_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.optitype_data, "multiqc_optitype")

        # Create OptiType overview table
        self.add_summary_metrics()

        # Create bar graph of subtype sample counts
        self.summary_barplot()

    def add_summary_metrics(self):
        """Take the parsed entries from OptiType and add them to the main plot"""
        # A1, A2, B1, B2, C1, C2, Reads, Objective
        headers = {
            "A1": {
                "title": "HLA-A1",
                "description": "First HLA-A allele",
                "scale": False,
            },
            "A2": {
                "title": "HLA-A2",
                "description": "Second HLA-A allele",
                "scale": False,
                "hidden": True,
            },
            "B1": {
                "title": "HLA-B1",
                "description": "First HLA-B allele",
                "scale": False,
            },
            "B2": {
                "title": "HLA-B2",
                "description": "Second HLA-B allele",
                "scale": False,
                "hidden": True,
            },
            "C1": {
                "title": "HLA-C1",
                "description": "First HLA-C allele",
                "scale": False,
            },
            "C2": {
                "title": "HLA-C2",
                "description": "Second HLA-C allele",
                "scale": False,
                "hidden": True,
            },
            "Reads": {
                "title": "Reads",
                "description": "Number of reads covering the HLA",
                "scale": "YlGnBu",
                "hidden": True,
                "format": "{:,.0f}",
            },
            "Objective": {
                "title": "Objective Score",
                "description": "Score of the objective function for the prediction.",
                "scale": "BuGn",
                "hidden": True,
            },
        }
        self.general_stats_addcols(self.optitype_data, headers)

    def summary_barplot(self):
        """Make a bar plot showing the number of samples for each allele"""

        alleles = ["A1", "A2", "B1", "B2", "C1", "C2"]
        data = {al: {} for al in alleles}
        for s_data in self.optitype_data.values():
            for al in alleles:
                if s_data[al] not in data[al]:
                    data[al][s_data[al]] = 0
                data[al][s_data[al]] += 1

        pconfig = {
            "id": "optitype_summary_plot",
            "title": "Optitype: Summary of alleles",
            "cpswitch": False,
            "ylab": "# Samples",
            "yDecimals": False,
            "use_legend": False,
        }

        self.add_section(
            name="Summary of alleles",
            description="Number of samples sharing the same allele.",
            anchor="optitype_summary",
            plot=bargraph.plot(data, pconfig=pconfig),
        )



""" MultiQC module to parse output from leeHom """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    leeHom module class, parses stderr logs.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="leeHom",
            anchor="leehom",
            href="https://github.com/grenaud/leeHom",
            info="is a program for the Bayesian reconstruction of ancient DNA",
            doi="10.1093/nar/gku699",
        )

        # Find and load any leeHom reports
        self.leehom_data = dict()

        for f in self.find_log_files("leehom", filehandles=True):
            parsed_data = self.parse_leehom_logs(f)
            if parsed_data is not None and len(parsed_data) > 0:
                self.leehom_data[f["s_name"]] = parsed_data
                self.add_data_source(f, f["s_name"])

                # Superfluous function call to confirm that it is used in this module
                # Replace None with actual version if it is available
                self.add_software_version(None, f["s_name"])

        # Filter to strip out ignored sample names
        self.leehom_data = self.ignore_samples(self.leehom_data)

        if len(self.leehom_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.leehom_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.leehom_data, "multiqc_leehom")

        # Basic Stats Table
        self.leehom_general_stats_table()

    def parse_leehom_logs(self, f):
        """Go through log file looking for leehom output"""
        regexes = {
            "total": r"Total reads[\s\:]+(\d+)",
            "merged_trimming": r"Merged \(trimming\)\s+(\d+)",
            "merged_overlap": r"Merged \(overlap\)\s+(\d+)",
            "kept": r"Kept PE/SR\s+(\d+)",
            "trimmed": r"Trimmed SR\s+(\d+)",
            "adapter_dimers_chimeras": r"Adapter dimers/chimeras\s+(\d+)",
            "failed_key": r"Failed Key\s+(\d+)",
        }
        parsed_data = dict()
        for line in f["f"]:
            # Search regexes for overview stats
            for k, r in regexes.items():
                match = re.search(r, line)
                if match:
                    parsed_data[k] = int(match.group(1))
        return parsed_data

    def leehom_general_stats_table(self):
        """Take the parsed stats from the leeHom report and add it to the
        basic stats table at the top of the report"""

        headers = {}
        headers["merged_trimming"] = {
            "title": f"{config.read_count_prefix} Merged (Trimming)",
            "description": f"Merged clusters from trimming ({config.read_count_desc})",
            "min": 0,
            "scale": "PuRd",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
        }
        headers["merged_overlap"] = {
            "title": f"{config.read_count_prefix} Merged (Overlap)",
            "description": f"Merged clusters from overlapping reads ({config.read_count_desc})",
            "min": 0,
            "scale": "PuRd",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
        }
        self.general_stats_addcols(self.leehom_data, headers)



#!/usr/bin/env python

""" MultiQC Utility functions, used in a variety of places. """


import io
import json
import logging
import os
import shutil
import sys
import time
import datetime
from typing import Dict, List, Union

import yaml

from . import config

log = logging.getLogger(__name__)


def robust_rmtree(path, logger=None, max_retries=10):
    """Robustly tries to delete paths.
    Retries several times (with increasing delays) if an OSError
    occurs.  If the final attempt fails, the Exception is propagated
    to the caller.
    """

    for i in range(max_retries):
        try:
            shutil.rmtree(path)
            return
        except OSError:
            if logger:
                logger.info(f"Unable to remove path: {path}")
                logger.info(f"Retrying after {i**2} seconds")
            else:
                print(f"Unable to remove path: {path}", file=sys.stderr)
                print(f"Retrying after {i**2} seconds", file=sys.stderr)
            time.sleep(i**2)

    # Final attempt, pass any Exceptions up to caller.
    shutil.rmtree(path)


def write_data_file(
    data: Union[Dict[str, Union[Dict, List]], List[Dict]],
    fn: str,
    sort_cols=False,
    data_format=None,
):
    """
    Write a data file to the report directory. Will not do anything
    if config.data_dir is not set.
    :param: data - either: a 2D dict, first key sample name (row header),
        second key field (column header); a list of dicts; or a list of lists
    :param: fn - Desired filename. Directory will be prepended automatically.
    :param: sort_cols - Sort columns alphabetically
    :param: data_format - Output format. Defaults to config.data_format (usually tsv)
    :return: None
    """

    if config.data_dir is None:
        return

    # Get data format from config
    if data_format is None:
        data_format = config.data_format

    # JSON encoder class to handle lambda functions
    class MQCJSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if callable(obj):
                # noinspection PyBroadException
                try:
                    return obj(1)
                except Exception:
                    return None
            return json.JSONEncoder.default(self, obj)

    body = None
    # Some metrics can't be coerced to tab-separated output, test and handle exceptions
    if data_format in ["tsv", "csv"]:
        sep = "\t" if data_format == "tsv" else ","
        # Attempt to reshape data to tsv
        # noinspection PyBroadException
        try:
            # Get all headers from the data, except if data is a dictionary (i.e. has >1 dimensions)
            headers = []
            rows = []

            for d in data.values() if isinstance(data, dict) else data:
                if not d or (isinstance(d, list) and isinstance(d[0], dict)):
                    continue
                if isinstance(d, dict):
                    for h in d.keys():
                        if h not in headers:
                            headers.append(h)
            if headers:
                if sort_cols:
                    headers = sorted(headers)
                headers_str = [str(item) for item in headers]
                if isinstance(data, dict):
                    # Add Sample header as a first element
                    headers_str.insert(0, "Sample")
                rows.append(sep.join(headers_str))

            # The rest of the rows
            for key, d in sorted(data.items()) if isinstance(data, dict) else enumerate(data):
                # Make a list starting with the sample name, then each field in order of the header cols
                if headers:
                    line = [str(d.get(h, "")) for h in headers]
                else:
                    line = [
                        str(item)
                        for item in (d.values() if isinstance(d, dict) else (d if isinstance(d, list) else [d]))
                    ]
                if isinstance(data, dict):
                    # Add Sample header as a first element
                    line.insert(0, str(key))
                rows.append(sep.join(line))
            body = "\n".join(rows)

        except Exception as e:
            if config.development:
                raise
            data_format = "yaml"
            log.debug(f"{fn} could not be saved as tsv/csv, falling back to YAML. {e}")

    # Add relevant file extension to filename, save file.
    fn = f"{fn}.{config.data_format_extensions[data_format]}"
    fpath = os.path.join(config.data_dir, fn)
    with io.open(fpath, "w", encoding="utf-8") as f:
        if data_format == "json":
            jsonstr = json.dumps(data, indent=4, cls=MQCJSONEncoder, ensure_ascii=False)
            print(jsonstr.encode("utf-8", "ignore").decode("utf-8"), file=f)
        elif data_format == "yaml":
            yaml.dump(data, f, default_flow_style=False)
        elif body:
            # Default - tab separated output
            print(body.encode("utf-8", "ignore").decode("utf-8"), file=f)
    log.debug(f"Wrote data file {fn}")


def force_term_colors():
    """
    Check if any environment variables are set to force Rich to use coloured output
    """
    if os.getenv("GITHUB_ACTIONS") or os.getenv("FORCE_COLOR") or os.getenv("PY_COLORS"):
        return True
    return None


def strtobool(val) -> bool:
    """
    Replaces deprecated https://docs.python.org/3.9/distutils/apiref.html#distutils.util.strtobool
    The deprecation recommendation is to re-implement the function https://peps.python.org/pep-0632/

    ------------------------------------------------------------

    Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val_str = str(val).lower()
    if val_str in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val_str in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"invalid truth value {val!r}")


def choose_emoji():
    """Choose an emoji to use in the report header."""
    # NB: We haven't parsed the config yet, so can't disable via config
    today = datetime.date.today()
    emojis = {
        "bottle_with_popping_cork": (1, 1, 1, 5),  # New Year's Day
        "rose": (2, 14, 0, 0),  # Valentine's Day
        "four_leaf_clover": (3, 17, 0, 0),  # St Patrick's Day
        "globe_showing_asia-australia": (4, 22, 0, 0),  # Earth Day
        "jack-o-lantern": (10, 31, 5, 0),  # Halloween
        "santa": (12, 25, 0, 0),  # Christmas Day
        "christmas_tree": (12, 25, 7, 7),  # Christmas
    }
    for emoji, (month, day, days_before, days_after) in emojis.items():
        special_date = datetime.date(today.year, month, day)
        date_range_start = special_date - datetime.timedelta(days=days_before)
        date_range_end = special_date + datetime.timedelta(days=days_after)
        if date_range_start <= today <= date_range_end:
            return emoji
    return "mag"


# Custom encoder to handle lambda functions
class MQCJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if callable(obj):
            try:
                return obj(1)
            except Exception:
                return None
        return json.JSONEncoder.default(self, obj)


def multiqc_dump_json(report):
    """
    Export the parsed data in memory to a JSON file.
    Used for MegaQC and other data export.
    WARNING: May be depreciated and removed in future versions.
    """
    exported_data = dict()
    export_vars = {
        "report": [
            "data_sources",
            "general_stats_data",
            "general_stats_headers",
            "multiqc_command",
            "plot_data",
            "saved_raw_data",
        ],
        "config": [
            "analysis_dir",
            "creation_date",
            "git_hash",
            "intro_text",
            "report_comment",
            "report_header_info",
            "script_path",
            "short_version",
            "subtitle",
            "title",
            "version",
            "output_dir",
        ],
    }
    for s in export_vars:
        for k in export_vars[s]:
            try:
                d = None
                if s == "config":
                    d = {f"{s}_{k}": getattr(config, k)}
                elif s == "report":
                    d = {f"{s}_{k}": getattr(report, k)}
                if d:
                    json.dumps(d, cls=MQCJSONEncoder, ensure_ascii=False)  # Test that exporting to JSON works
                    exported_data.update(d)
            except (TypeError, KeyError, AttributeError) as e:
                log.warning(f"Couldn't export data key '{s}.{k}': {e}")
        # Get the absolute paths of analysis directories
        exported_data["config_analysis_dir_abs"] = list()
        for d in exported_data.get("config_analysis_dir", []):
            try:
                exported_data["config_analysis_dir_abs"].append(os.path.abspath(d))
            except Exception:
                pass
    return exported_data



""" MultiQC submodule to parse output from Picard VariantCallingMetrics """

import logging

from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard VariantCallingMetrics reports and process their data"""

    # get data
    data_by_sample = collect_data(module)

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    derive_data(data_by_sample)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, "multiqc_picard_variantCalling")

    headers = {
        "DBSNP_TITV": {
            "title": "TiTV ratio (known)",
            "description": "The Transition/Transversion ratio of the passing bi-allelic SNP calls made at SNP-database sites.",
            "min": 0,
            "scale": "Blues",
            "shared_key": "titv_ratio",
        },
        "NOVEL_TITV": {
            "title": "TiTV ratio (novel)",
            "description": "The Transition/Transversion ratio of the passing bi-allelic SNP calls made at non-SNP-database sites.",
            "min": 0,
            "scale": "Blues",
            "shared_key": "titv_ratio",
        },
        "DBSNP_INS_DEL_RATIO": {
            "title": "InDel ratio (known)",
            "description": "The Insertion / Deletion ratio of the passing bi-allelic SNP calls made at SNP-database sites.",
            "min": 0,
            "scale": "Greens",
            "shared_key": "indel_ratio",
            "hidden": True,
        },
        "NOVEL_INS_DEL_RATIO": {
            "title": "InDel ratio (novel)",
            "description": "The Insertion / Deletion ratio of the passing bi-allelic SNP calls made at non-SNP-database sites.",
            "min": 0,
            "scale": "Greens",
            "shared_key": "indel_ratio",
            "hidden": True,
        },
        "total_called_variants_known": {
            "title": "Called Variants (known)",
            "description": "Total counts of variants in SNP-database sites.",
            "shared_key": "variant_count",
            "min": 0,
            "format": "{0:,.0f}",
            "hidden": True,
        },
        "total_called_variants_novel": {
            "title": "Called Variants (novel)",
            "description": "Total counts of variants in non-SNP-database sites.",
            "shared_key": "variant_count",
            "min": 0,
            "format": "{0:,.0f}",
            "hidden": True,
        },
    }
    module.general_stats_addcols(data_by_sample, headers, namespace="VariantCallingMetrics")

    # Variant Counts Bargraph
    module.add_section(
        name="Variant Types",
        anchor="picard-variants-types",
        description="Variants that have been called, looking at variant types. Optionally filtered on label.",
        helptext="""
        Only passing variants are shown (i.e. non-filtered).\n
        SNPs are bi-allelic.\n
        Complex InDels are both an insertion and a deletion.
        """,
        plot=compare_variant_type_plot(data_by_sample),
    )

    # Variant Counts Table
    module.add_section(
        name="Variant Labels",
        anchor="picard-variants-labels",
        description="Variants that have been called, comparing with known variant sites.",
        helptext="""
        Only passing variants are shown (i.e. non-filtered).\n
        Variants contain bi-allelic SNPs, multi-allelic SNPs, simple and complex inserts and deletions.
        """,
        plot=compare_variants_label_plot(data_by_sample),
    )

    return len(data_by_sample)


def collect_data(module):
    """Find Picard VariantCallingMetrics reports and parse their data"""

    data = dict()
    for f in module.find_log_files("picard/variant_calling_metrics", filehandles=True):
        s_name = None
        for header, value in table_in(f["f"], pre_header_string="## METRICS CLASS"):
            if header == "SAMPLE_ALIAS":
                s_name = value
                if s_name in data:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                data[s_name] = dict()
            else:
                data[s_name][header] = value
    return data


def table_in(filehandle, pre_header_string):
    """Generator that assumes a table starts the line after a given string"""

    in_histogram = False
    next_is_header = False
    headers = list()
    for line in stripped(filehandle):
        if not in_histogram and line.startswith(pre_header_string):
            in_histogram = True
            next_is_header = True
        elif in_histogram and next_is_header:
            next_is_header = False
            headers = line.split("\t")
        elif in_histogram:
            values = line.split("\t")
            if values != [""]:
                for couple in zip(headers, values):
                    yield couple


def derive_data(data):
    """Based on the data derive additional data"""

    for s_name, values in data.items():
        # setup holding variable

        # Sum all variants that have been called
        total_called_variants = 0
        for value_name in ["TOTAL_SNPS", "TOTAL_COMPLEX_INDELS", "TOTAL_MULTIALLELIC_SNPS", "TOTAL_INDELS"]:
            total_called_variants = total_called_variants + int(values[value_name])
        values["total_called_variants"] = total_called_variants

        # Sum all variants that have been called and are known
        total_called_variants_known = 0
        for value_name in ["NUM_IN_DB_SNP", "NUM_IN_DB_SNP_COMPLEX_INDELS", "NUM_IN_DB_SNP_MULTIALLELIC"]:
            total_called_variants_known = total_called_variants_known + int(values[value_name])
        total_called_variants_known = (
            total_called_variants_known + int(values["TOTAL_INDELS"]) - int(values["NOVEL_INDELS"])
        )
        values["total_called_variants_known"] = total_called_variants_known

        # Extrapolate the total novel variants
        values["total_called_variants_novel"] = total_called_variants - total_called_variants_known


def stripped(iterator):
    """Generator to strip string of whitespace"""
    for item in iterator:
        yield item.strip()


def compare_variant_type_plot(data):
    """Return HTML for the Variant Counts barplot"""
    keys = dict()
    keys["snps"] = {"name": "SNPs", "color": "#7cb5ec"}
    keys["indels"] = {"name": "InDels", "color": "#90ed7d"}
    keys["multiallelic_snps"] = {"name": "multi-allelic SNP", "color": "orange"}
    keys["complex_indels"] = {"name": "Complex InDels", "color": "#8085e9"}

    total_variants = dict()
    known_variants = dict()
    novel_variants = dict()
    for s_name, values in data.items():
        total_variants[s_name] = {
            "snps": values["TOTAL_SNPS"],
            "indels": values["TOTAL_INDELS"],
            "multiallelic_snps": values["TOTAL_MULTIALLELIC_SNPS"],
            "complex_indels": values["TOTAL_COMPLEX_INDELS"],
        }

        known_variants[s_name] = {
            "snps": values["NUM_IN_DB_SNP"],
            "indels": int(values["TOTAL_INDELS"]) - int(values["NOVEL_INDELS"]),
            "multiallelic_snps": values["NUM_IN_DB_SNP_MULTIALLELIC"],
            "complex_indels": values["NUM_IN_DB_SNP_COMPLEX_INDELS"],
        }

        novel_variants[s_name] = {
            "snps": values["NOVEL_SNPS"],
            "indels": values["NOVEL_INDELS"],
            "multiallelic_snps": int(values["TOTAL_MULTIALLELIC_SNPS"]) - int(values["NUM_IN_DB_SNP_MULTIALLELIC"]),
            "complex_indels": int(values["TOTAL_COMPLEX_INDELS"]) - int(values["NUM_IN_DB_SNP_COMPLEX_INDELS"]),
        }

    plot_conf = {
        "id": "picard_variantCallingMetrics_variant_type",
        "title": "Picard: Variants Called",
        "ylab": "Counts of Variants",
        "hide_zero_cats": False,
        "data_labels": [{"name": "Total"}, {"name": "Known"}, {"name": "Novel"}],
    }
    return bargraph.plot(
        data=[total_variants, known_variants, novel_variants], cats=[keys, keys, keys], pconfig=plot_conf
    )


def compare_variants_label_plot(data):
    """Return HTML for the Compare variants plot"""
    keys = dict()
    keys["total_called_variants_known"] = {"name": "Known Variants"}
    keys["total_called_variants_novel"] = {"name": "Novel Variants"}

    pconfig = {
        "id": "picard_variantCallingMetrics_variant_label",
        "title": "Picard: Variants Called",
        "ylab": "Counts of Variants",
    }

    return bargraph.plot(data, cats=keys, pconfig=pconfig)


def merge_two_dicts(x, y):
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z
